#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Tested with the following GPUs: 3090
import pytest

from smartg.smartg import Smartg, Sensor, LambSurface, Albedo_cst
from smartg.atmosphere import AtmAFGL
import pandas as pd
import numpy as np

from smartg.iprt.iprt import convert_SGout_to_IPRTout, select_and_plot_polar_iprt, compute_deltam, seclect_iprt_IQUV, plot_iprt_radiances, groupIQUV
from smartg.libATM3D import read_cld_nth_cte
from smartg.tools.phase import calc_iphase
from luts.luts import LUT
from smartg.config import DIR_AUXDATA

from smartg import conftest

import matplotlib.pyplot as plt
import matplotlib.image as mpimg

from tempfile import TemporaryDirectory
from pathlib import Path

import logging

# ***************************** Global variable(s) ******************************
SEED = -1
STDFAC = 4
ROOTPATH = Path(__file__).resolve().parent.parent
# *******************************************************************************

# *********************************** logging ***********************************
# Create log file
log_dir = ROOTPATH / "tests" / "logs"
log_dir.mkdir(parents=True, exist_ok=True)

# Create a named logger
logger = logging.getLogger('test_phaseA')
logger.setLevel(logging.INFO)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.ERROR)

# Set the formatter for the console handler
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s',
datefmt='%m/%d/%Y %I:%M:%S%p')
console_handler.setFormatter(formatter)

# Add the console handler to the logger
logger.addHandler(console_handler)

# Create a file handler
file_handler = logging.FileHandler(ROOTPATH / 'tests' / 'logs' / 'iprt_phaseA.log', mode='w')
file_handler.setLevel(logging.INFO)

# Set the formatter for the file handler
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%m/%d/%Y %I:%M:%S%p')
file_handler.setFormatter(formatter)

# Add the file handler to the logger
logger.addHandler(file_handler)
# *******************************************************************************

@pytest.fixture(scope='module')
def S1DF():
    '''
    Forward compilation in 1D
    '''
    return Smartg(alt_pp=True, back=False, double=True, bias=True)

@pytest.fixture(scope='module')
def S1DB():
    '''
    Backward compilation in 1D
    '''
    return Smartg(alt_pp=True, back=True, double=True, bias=True)


def test_A1(request, S1DF, S1DB):
    print(("=== Test A1"))
    mol_sca = np.array([0., 0.5])[None,:]
    mol_abs = np.array([0., 0.])[None,:]
    z       = np.array([1., 0.])
    atm     = AtmAFGL('afglt', grid=z, prof_ray= mol_sca, prof_abs=mol_abs).calc(550.)
    surf    = None

    # *************************** DEPOL = 0 *************************** 
    SZA = 0.
    SAA = 65.
    PHI_0 = 180.-SAA # To follow MYSTIC convention
    le     = {'th_deg':np.array([SZA]), 'phi_deg':np.array([PHI_0]), 'count_level':np.array([0])}

    # BOA radiances
    VZAMIN = 0.
    VZAMAX = 80.
    VZAINC = 5.
    VZA = np.arange(VZAMIN, VZAMAX+VZAINC, VZAINC)

    # VAA from 0. to 180.
    VAAMIN = 0.
    VAAMAX = 360.
    VAAINC = 5.
    VAA = np.arange(VAAMIN, VAAMAX+VAAINC, VAAINC)

    lsensors = []
    NBVZA = len(VZA)
    NBVAA = len(VAA)
    NBDIR = round(NBVAA*NBVZA)
    for iza, za in enumerate(VZA):
        for iaa, aa in enumerate(VAA):
            PHI = -aa+180
            lsensors.append(Sensor(POSZ=np.min(z), THDEG=za, PHDEG=PHI, LOC='ATMOS'))

    mA1B = S1DB.run(wl=550., NBPHOTONS=1e7*NBDIR, NBLOOP=1e7, atm=atm, sensor=lsensors, OUTPUT_LAYERS=0,
                    le=le, surf=surf, XBLOCK = 64, XGRID = 1024, BEER=1, DEPO=0.0, stdev=True, progress=True)

    mA1B = mA1B.dropaxis('Azimuth angles', 'Zenith angles')
    mA1B.add_axis('Azimuth angles', -VAA+180.)
    mA1B.add_axis('Zenith angles', VZA)

    for name in mA1B.datasets():
        if 'sensor index' in mA1B[name].names:
            mat_tmp = np.swapaxes(mA1B[name][:].reshape(len(VZA), len(VAA)), 0, 1)
            attrs_tmp = mA1B[name].attrs
            mA1B.rm_lut(name)
            mA1B.add_dataset(name, mat_tmp, ['Azimuth angles', 'Zenith angles'], attrs=attrs_tmp)

    mA1B_boa_dep0 = mA1B.dropaxis('sensor index')

    # TOA radiances
    # We use the previous VAA
    VZAMIN = 100.
    VZAMAX = 180.
    VZAINC = 5.
    VZA = np.arange(VZAMIN, VZAMAX+VZAINC, VZAINC)

    lsensors = []
    NBVZA = len(VZA)
    NBVAA = len(VAA)
    NBDIR = round(NBVAA*NBVZA)
    for iza, za in enumerate(VZA):
        for iaa, aa in enumerate(VAA):
            PHI = -aa+180
            lsensors.append(Sensor(POSZ=np.max(z), THDEG=za, PHDEG=PHI, LOC='ATMOS'))

    mA1B = S1DB.run(wl=550., NBPHOTONS=1e7*NBDIR, NBLOOP=1e7, atm=atm, sensor=lsensors, OUTPUT_LAYERS=0,
                    le=le, surf=surf, XBLOCK = 64, XGRID = 1024, BEER=1, DEPO=0.0, stdev=True, progress=True)

    mA1B = mA1B.dropaxis('Azimuth angles', 'Zenith angles')
    mA1B.add_axis('Azimuth angles', -VAA+180.)
    mA1B.add_axis('Zenith angles', VZA)

    for name in mA1B.datasets():
        if 'sensor index' in mA1B[name].names:
            mat_tmp = np.swapaxes(mA1B[name][:].reshape(len(VZA), len(VAA)), 0, 1)
            attrs_tmp = mA1B[name].attrs
            mA1B.rm_lut(name)
            mA1B.add_dataset(name, mat_tmp, ['Azimuth angles', 'Zenith angles'], attrs=attrs_tmp)

    mA1B_toa_dep0 = mA1B.dropaxis('sensor index')
    # *****************************************************************

    # ************************* DEPOL = 0.03 ************************** 
    # We use the previous VAA and VZA
    # SMART-G Forward TH and PHI using local estimate (anticlockwise) conversion with VZA and VAA MYSTIC (clockwise)
    TH  = 180.-VZA
    PHI = -VAA
    TH[TH==0] = 1e-6  # avoid problem due to special case of 0
    le     = {'th_deg':TH, 'phi_deg':PHI}#, 'zip':True}
    SZA = 30.0
    SAA = 0.
    PHI_0 = 180.-SAA # SMART-G anticlockwise converted to be consistent with MYSTIC
    mA1F_dep003 = S1DF.run(THVDEG=SZA, PHVDEG=PHI_0, wl=550., NBPHOTONS=1e7, NBLOOP=1e5, atm=atm, OUTPUT_LAYERS=int(7),
                        le=le, surf=surf, XBLOCK = 64, XGRID = 1024, BEER=1, DEPO=0.03, stdev=True)

    # ************************* DEPOL = 0.1 **************************
    # We use the previous VAA, VZA and le
    SZA = 30.0
    SAA = 65.0
    PHI_0 = 180.-SAA # SMART-G anticlockwise converted to be consistent with MYSTIC
    mA1F_dep01 = S1DF.run(THVDEG=SZA, PHVDEG=PHI_0, wl=550., NBPHOTONS=1e7, NBLOOP=1e5, atm=atm, OUTPUT_LAYERS=int(7),
                        le=le, surf=surf, XBLOCK = 64, XGRID = 1024, BEER=1, DEPO=0.1, stdev=True)
    # *****************************************************************

    with TemporaryDirectory() as tmpdir:
        # === Convert smartg output to iprt ascii output format (Forward, U must be multiplied by -1)
        tmp_file_a1 = Path(tmpdir)/'a1.dat'

        VZA_boa_dep0 = mA1B_boa_dep0.axes['Zenith angles']
        VAA_boa_dep0 = 180.-mA1B_boa_dep0.axes['Azimuth angles']
        VZA_toa_dep0 = mA1B_toa_dep0.axes['Zenith angles']
        VAA_toa_dep0 = 180.-mA1B_toa_dep0.axes['Azimuth angles']

        VZA_dep003 = 180.-mA1F_dep003.axes['Zenith angles']
        VAA_dep003 = -mA1F_dep003.axes['Azimuth angles']
        VZA_dep01 = 180.-mA1F_dep01.axes['Zenith angles']
        VAA_dep01 = -mA1F_dep01.axes['Azimuth angles']

        # convert
        convert_SGout_to_IPRTout(lm=[mA1B_boa_dep0, mA1B_toa_dep0, mA1F_dep003, mA1F_dep003, mA1F_dep01, mA1F_dep01], lU_sign=[1., 1, -1, -1, -1, -1], case_name="A1",
                                ldepol=[0., 0., 0.03, 0.03, 0.1, 0.1], lalt=[0., 1., 0., 1., 0., 1.], lSZA=[0., 0., 30., 30., 30., 30.], lSAA=[65., 65., 0., 0., 65., 65.],
                                lVZA=[VZA_boa_dep0, VZA_toa_dep0, 180.-VZA_dep003, VZA_dep003, 180.-VZA_dep01, VZA_dep01], lVAA=[VAA_boa_dep0, VAA_toa_dep0, VAA_dep003, VAA_dep003, VAA_dep01, VAA_dep01],
                                file_name=tmp_file_a1, output_layer=['_up (TOA)', '_up (TOA)', '_down (0+)', '_up (TOA)', '_down (0+)', '_up (TOA)'])
        
        smartg_a1 = pd.read_csv(tmp_file_a1, header=None, sep=r'\s+', dtype=float, comment="#").values
        mystic_a1 = pd.read_csv(DIR_AUXDATA / 'IPRT' / 'phaseA' / 'mystic_res' / 'iprt_case_a1_mystic.dat',
                                header=None, sep=r'\s+', dtype=float, comment="#").values
        avoidP=False

        lDEP   = [0., 0.03, 0.1]
        lSZA   = [0., 30., 30.]
        lSAA   = [65., 0., 65.]
        lALT   = [0., 1.]
        lINVTH = [False, True]
        # ============ 0km of altitude
        for isim in range (0, 3):
            imgs = []
            for ialt in range (0, 2):
                title = f"IPRT case A1 - depol = {lDEP[isim]} - SZA = {lSZA[isim]:.0f} - SAA = {lSAA[isim]:.0f} - {lALT[ialt]:.0f}km - SMARTG"
                tmp_filename = f'a1_dep{lDEP[isim]:.0e}_{lALT[ialt]:.0f}km_smartg.png'
                I_smartg, Q_smartg, U_smartg, V_smartg = select_and_plot_polar_iprt(smartg_a1, z_alti=lALT[ialt], depol=lDEP[isim], title=title, change_U_sign=True, inv_thetas=lINVTH[ialt], sym=False, outputIQUV=True, avoid_plot=avoidP, save_fig=Path(tmpdir)/tmp_filename)
                imgs.append(mpimg.imread(Path(tmpdir)/tmp_filename))

                tmp_filename = f'a1_dep{lDEP[isim]}_{lALT[ialt]:.0f}km_mystic.png'
                title = f"IPRT case A1 - depol = {lDEP[isim]} - SZA = {lSZA[isim]:.0f} - SAA = {lSAA[isim]:.0f}  - {lALT[ialt]:.0f}km - MYSTIC"
                I_mystic, Q_mystic, U_mystic, V_mystic = select_and_plot_polar_iprt(mystic_a1, z_alti=lALT[ialt], depol=lDEP[isim], title=title, change_U_sign=True, inv_thetas=lINVTH[ialt], sym=False, outputIQUV=True, avoid_plot=avoidP, save_fig=Path(tmpdir)/tmp_filename)
                imgs.append(mpimg.imread(Path(tmpdir)/tmp_filename))

                if (isim == 0 and ialt == 0):
                    IQUV_smartg_tot = groupIQUV(lI=[I_smartg], lQ=[Q_smartg], lU=[U_smartg], lV=[V_smartg])
                    IQUV_mystic_tot = groupIQUV(lI=[I_mystic], lQ=[Q_mystic], lU=[U_mystic], lV=[V_mystic])
                else:
                    IQUV_smartg_tot = np.concatenate((IQUV_smartg_tot, groupIQUV(lI=[I_smartg], lQ=[Q_smartg], lU=[U_smartg], lV=[V_smartg])), axis=1)
                    IQUV_mystic_tot = np.concatenate((IQUV_mystic_tot, groupIQUV(lI=[I_mystic], lQ=[Q_mystic], lU=[U_mystic], lV=[V_mystic])), axis=1)

                I_val = I_mystic-I_smartg
                Q_val = Q_mystic-Q_smartg
                U_val = U_mystic-U_smartg
                V_val = V_mystic-V_smartg
                maxI = max(np.abs(np.min(I_val)), np.abs(np.max(I_val)))
                maxQ = max(np.abs(np.min(Q_val)), np.abs(np.max(Q_val)))
                maxU = max(np.abs(np.min(U_val)), np.abs(np.max(U_val)))
                maxV = max(np.abs(np.min(V_val)), np.abs(np.max(V_val)))
                tmp_filename = f'a1_dep{lDEP[isim]}_0{lALT[ialt]:.0f}m_dif.png'
                title = f"IPRT case A1 - depol = {lDEP[isim]}  - SZA = {lSZA[isim]:.0f} - SAA = {lSAA[isim]:.0f} - {lALT[ialt]:.0f}km - dif (MYSTIC-SMARTG)"
                select_and_plot_polar_iprt(mystic_a1, z_alti=lALT[ialt], depol=lDEP[isim], title=title, forceIQUV=[I_val, Q_val, U_val, V_val], maxI=maxI, maxQ=maxQ, maxU=maxU, maxV=maxV, cmapI='RdBu_r', avoid_plot=avoidP, save_fig=Path(tmpdir)/tmp_filename)
                imgs.append(mpimg.imread(Path(tmpdir)/tmp_filename))

            plt.close('all')
            fig, axs = plt.subplots(6,1, figsize=(12, 24))
            for i in range (0, 6):
                axs[i].axis('off')
                axs[i].imshow(imgs[i])
            fig.tight_layout()
            conftest.savefig(request, bbox_inches='tight')


    # === Compute the delta_m values and analyse them with the previous saved validated ones
    # SMARTG ref results
    smartg_a1_ref = pd.read_csv(DIR_AUXDATA / 'IPRT' / 'phaseA' / 'smartg_ref_res' / 'iprt_output_format' / 'iprt_case_a1_smartg_ref.dat',
                                header=None, sep=r'\s+', dtype=float, comment="#").values
    for isim in range (0, 3):
        for ialt in range (0, 2):
            I_smartg_ref, Q_smartg_ref, U_smartg_ref, V_smartg_ref, I_smartg_std_ref, Q_smartg_std_ref, U_smartg_std_ref, V_smartg_std_ref \
                = select_and_plot_polar_iprt(smartg_a1_ref, z_alti=lALT[ialt], depol=lDEP[isim], change_U_sign=True, inv_thetas=lINVTH[ialt], sym=False, outputIQUV=True, outputIQUVstd=True, avoid_plot=True)

            if (isim == 0 and ialt == 0):
                IQUV_smartg_ref_tot = groupIQUV(lI=[I_smartg_ref], lQ=[Q_smartg_ref], lU=[U_smartg_ref], lV=[V_smartg_ref])
                IQUV_smartg_std_ref_tot = groupIQUV(lI=[I_smartg_std_ref], lQ=[Q_smartg_std_ref], lU=[U_smartg_std_ref], lV=[V_smartg_std_ref])
            else:
                IQUV_smartg_ref_tot = np.concatenate((IQUV_smartg_ref_tot, groupIQUV(lI=[I_smartg_ref], lQ=[Q_smartg_ref], lU=[U_smartg_ref], lV=[V_smartg_ref])), axis=1)
                IQUV_smartg_std_ref_tot = np.concatenate((IQUV_smartg_std_ref_tot, groupIQUV(lI=[I_smartg_std_ref], lQ=[Q_smartg_std_ref], lU=[U_smartg_std_ref], lV=[V_smartg_std_ref])), axis=1)

    # Compute the delta_m values from the ref smartg results
    delta_m_ref = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_ref_tot, print_res=False)
    logger.info(f'A1 - I={delta_m_ref[0]:.3f}; Q={delta_m_ref[1]:.3f}; U={delta_m_ref[2]:.3f}; V={delta_m_ref[3]:.3f} - ref delta_m:')

    # Compute the delta_m values from the ref smartg results +- err
    delta_m_ref_P = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_ref_tot+STDFAC*IQUV_smartg_std_ref_tot, print_res=False)
    delta_m_ref_M = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_ref_tot-STDFAC*IQUV_smartg_std_ref_tot, print_res=False)

    # Compute the delta_m values from the smartg test results
    delta_m = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_tot, print_res=False)
    logger.info(f'A1 - I={delta_m[0]:.3f}; Q={delta_m[1]:.3f}; U={delta_m[2]:.3f}; V={delta_m[3]:.3f} - calculated delta_m')

    # Check if the the test is ok by comparing smartg ref and smartg test
    IQUV_name = ['I', 'Q', 'U', 'V']
    for istk, stk in enumerate(IQUV_name):
        maxVal = max(delta_m_ref[istk], delta_m_ref_P[istk], delta_m_ref_M[istk])
        assert not ( delta_m[istk] > maxVal ), f'Problem with {stk} values, get {delta_m[istk]:.5f}. {stk} must be < to {maxVal:.5f}'


def test_A2(request, S1DF):
    print("=== Test A2:")
    # === Atmosphere profil
    mol_sca = np.array([0., 0.1])[None,:]
    mol_abs = np.array([0., 0.])[None,:]
    z       = np.array([1., 0.])
    atm     = AtmAFGL('afglt', grid=z, prof_ray=mol_sca, prof_abs=mol_abs).calc(550.)
    surf    = LambSurface(ALB=Albedo_cst(0.3))

    # === Illumination conditions
    VZAMIN = 100.
    VZAMAX = 180.
    VZAINC = 5.
    VZA = np.arange(VZAMIN, VZAMAX+VZAINC, VZAINC)

    VAAMIN = 0.
    VAAMAX = 180.
    VAAINC = 5.
    VAA = np.arange(VAAMIN, VAAMAX+VAAINC, VAAINC)

    # SMART-G Forward TH and PHI using local estimate (anticlockwise) conversion with VZA and VAA MYSTIC (clockwise)
    TH  = 180.-VZA
    PHI = -VAA
    TH[TH==0] = 1e-6  # avoid problem due to special case of 0
    le     = {'th_deg':TH, 'phi_deg':PHI}

    SZA = 50.
    SAA = 0.
    PHI_0 = 180.-SAA # SMART-G anticlockwise converted to be consistent with MYSTIC

    # === Simulation
    mA2F = S1DF.run(THVDEG=SZA, PHVDEG=PHI_0, wl=550., NBPHOTONS=1e7, NBLOOP=1e6, atm=atm, OUTPUT_LAYERS=int(7),
                    le=le, surf=surf, XBLOCK = 64, XGRID = 1024, BEER=1, DEPO=0.03, stdev=True, SEED=SEED)

    with TemporaryDirectory() as tmpdir:
        # === Convert smartg output to iprt ascii output format (Forward, U must be multiplied by -1)
        tmp_file_a2 = Path(tmpdir)/'a2.dat'
        convert_SGout_to_IPRTout(lm=[mA2F, mA2F], lU_sign=[-1, -1], case_name="A2", ldepol=[0.03, 0.03], lalt=[0., 1.], lSZA=[50., 50.], lSAA=[0., 0.], lVZA=[180.-VZA, VZA],
                                lVAA=[VAA, VAA], file_name= tmp_file_a2, output_layer=['_down (0+)', '_up (TOA)'])
        
        # === Plot and comparison with MYSTIC (to save in the report)
        smartg_a2 = pd.read_csv(tmp_file_a2, header=None, sep=r'\s+', dtype=float, comment="#").values
        mystic_a2 = pd.read_csv(DIR_AUXDATA / 'IPRT' / 'phaseA' / 'mystic_res' / 'iprt_case_a2_mystic.dat',
                                header=None, sep=r'\s+', dtype=float, comment="#").values
        avoidP=False
        # 0km of altitude
        title = "IPRT case A2 - depol = 0.03 - 0km - SMARTG"
        I_smartg_0km, Q_smartg_0km, U_smartg_0km, V_smartg_0km = select_and_plot_polar_iprt(smartg_a2, 0., title=title, change_U_sign=True, sym=True, outputIQUV=True, avoid_plot=avoidP, save_fig=Path(tmpdir)/'a2_0km_smartg.png')

        title = "IPRT case A2 - depol = 0.03 - 0km - MYSTIC"
        I_mystic_0km, Q_mystic_0km, U_mystic_0km, V_mystic_0km = select_and_plot_polar_iprt(mystic_a2, 0., title=title, change_U_sign=True, sym=True, outputIQUV=True, avoid_plot=avoidP, save_fig=Path(tmpdir)/'a2_0km_mystic.png')

        I_val = I_mystic_0km-I_smartg_0km
        Q_val = Q_mystic_0km-Q_smartg_0km
        U_val = U_mystic_0km-U_smartg_0km
        V_val = V_mystic_0km-V_smartg_0km
        maxI = max(np.abs(np.min(I_val)), np.abs(np.max(I_val)))
        maxQ = max(np.abs(np.min(Q_val)), np.abs(np.max(Q_val)))
        maxU = max(np.abs(np.min(U_val)), np.abs(np.max(U_val)))
        maxV = max(np.abs(np.min(V_val)), np.abs(np.max(V_val)))
        title = "IPRT case A2 - depol = 0.03 - 0km - dif (MYSTIC-SMARTG)"
        select_and_plot_polar_iprt(mystic_a2, 0., title=title, forceIQUV=[I_val, Q_val, U_val, V_val], maxI=maxI, maxQ=maxQ, maxU=maxU, maxV=maxV, cmapI='RdBu_r', avoid_plot=avoidP, save_fig=Path(tmpdir)/'a2_0km_dif.png')

        imgs = []
        imgs.append(mpimg.imread(Path(tmpdir)/'a2_0km_smartg.png'))
        imgs.append(mpimg.imread(Path(tmpdir)/'a2_0km_mystic.png'))
        imgs.append(mpimg.imread(Path(tmpdir)/'a2_0km_dif.png'))

    plt.close('all')
    fig, axs = plt.subplots(3,1, figsize=(12,12))
    for i in range (0, 3):
        axs[i].axis('off')
        axs[i].imshow(imgs[i])
    fig.tight_layout()
    conftest.savefig(request, bbox_inches='tight')

    with TemporaryDirectory() as tmpdir:
        # 1km of altitude (TOA)
        title = "IPRT case A2 - depol = 0.03 - 1km - SMARTG"
        I_smartg_1km, Q_smartg_1km, U_smartg_1km, V_smartg_1km = select_and_plot_polar_iprt(smartg_a2, 1., title=title, change_U_sign=True, inv_thetas=True,sym=True, outputIQUV=True, avoid_plot=avoidP, save_fig=Path(tmpdir)/"a2_1km_smartg.png")

        title = "IPRT case A2 - depol = 0.03 - 1km - MYSTIC"
        I_mystic_1km, Q_mystic_1km, U_mystic_1km, V_mystic_1km = select_and_plot_polar_iprt(mystic_a2, 1., title=title, change_U_sign=True, inv_thetas=True, sym=True, outputIQUV=True, avoid_plot=avoidP, save_fig=Path(tmpdir)/"a2_1km_mystic.png")

        I_val = I_mystic_1km-I_smartg_1km
        Q_val = Q_mystic_1km-Q_smartg_1km
        U_val = U_mystic_1km-U_smartg_1km
        V_val = V_mystic_1km-V_smartg_1km
        maxI = max(np.abs(np.min(I_val)), np.abs(np.max(I_val)))
        maxQ = max(np.abs(np.min(Q_val)), np.abs(np.max(Q_val)))
        maxU = max(np.abs(np.min(U_val)), np.abs(np.max(U_val)))
        maxV = max(np.abs(np.min(V_val)), np.abs(np.max(V_val)))
        title = "IPRT case A2 - depol = 0.03 - 1km - dif (MYSTIC-SMARTG)"
        select_and_plot_polar_iprt(mystic_a2, 1., title=title, forceIQUV=[I_val, Q_val, U_val, V_val], maxI=maxI, maxQ=maxQ, maxU=maxU, maxV=maxV, cmapI='RdBu_r', avoid_plot=avoidP, save_fig=Path(tmpdir)/"a2_1km_dif.png")

        imgs = []
        imgs.append(mpimg.imread(Path(tmpdir)/'a2_1km_smartg.png'))
        imgs.append(mpimg.imread(Path(tmpdir)/'a2_1km_mystic.png'))
        imgs.append(mpimg.imread(Path(tmpdir)/'a2_1km_dif.png'))

    plt.close('all')
    fig, axs = plt.subplots(3,1, figsize=(12,12)) #12,8

    for i in range (0, 3):
        axs[i].axis('off')
        axs[i].imshow(imgs[i])
    fig.tight_layout()
    conftest.savefig(request, bbox_inches='tight')

    # === Compute the delta_m values and analyse them with the previous saved validated ones
    # MYSTIC and calculated SMART-G total IQUV
    IQUV_smartg_tot = groupIQUV(lI=[I_smartg_0km, I_smartg_1km], lQ=[Q_smartg_0km, Q_smartg_1km], lU=[U_smartg_0km, U_smartg_1km], lV=[V_smartg_0km, V_smartg_1km])
    IQUV_mystic_tot = groupIQUV(lI=[I_mystic_0km, I_mystic_1km], lQ=[Q_mystic_0km, Q_mystic_1km], lU=[U_mystic_0km, U_mystic_1km], lV=[V_mystic_0km, V_mystic_1km])

    # SMARTG ref results
    smartg_a2_ref = pd.read_csv(DIR_AUXDATA / 'IPRT' / 'phaseA' / 'smartg_ref_res' / 'iprt_output_format' / 'iprt_case_a2_smartg_ref.dat',
                                header=None, sep=r'\s+', dtype=float, comment="#").values
    I_smartg_0km_ref, Q_smartg_0km_ref, U_smartg_0km_ref, V_smartg_0km_ref, I_smartg_std_0km_ref, Q_smartg_std_0km_ref, U_smartg_std_0km_ref, V_smartg_std_0km_ref \
          = select_and_plot_polar_iprt(smartg_a2_ref, 0., change_U_sign=True, outputIQUV=True, outputIQUVstd=True, avoid_plot=True)
    I_smartg_1km_ref, Q_smartg_1km_ref, U_smartg_1km_ref, V_smartg_1km_ref, I_smartg_std_1km_ref, Q_smartg_std_1km_ref, U_smartg_std_1km_ref, V_smartg_std_1km_ref \
          = select_and_plot_polar_iprt(smartg_a2_ref, 1., change_U_sign=True, inv_thetas=True, outputIQUVstd=True, outputIQUV=True, avoid_plot=True)
    
    IQUV_smartg_ref_tot = groupIQUV(lI=[I_smartg_0km_ref, I_smartg_1km_ref], lQ=[Q_smartg_0km_ref, Q_smartg_1km_ref], lU=[U_smartg_0km_ref, U_smartg_1km_ref], lV=[V_smartg_0km_ref, V_smartg_1km_ref])
    IQUV_smartg_std_ref_tot = groupIQUV(lI=[I_smartg_std_0km_ref, I_smartg_std_1km_ref], lQ=[Q_smartg_std_0km_ref, Q_smartg_std_1km_ref],
                                        lU=[U_smartg_std_0km_ref, U_smartg_std_1km_ref], lV=[V_smartg_std_0km_ref, V_smartg_std_1km_ref])

    # Compute the delta_m values from the ref smartg results
    delta_m_ref = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_ref_tot, print_res=False)
    logger.info(f'A2 - I={delta_m_ref[0]:.3f}; Q={delta_m_ref[1]:.3f}; U={delta_m_ref[2]:.3f}; V={delta_m_ref[3]:.3f} - ref delta_m:')

    # Compute the delta_m values from the ref smartg results +- err
    delta_m_ref_P = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_ref_tot+STDFAC*IQUV_smartg_std_ref_tot, print_res=False)
    delta_m_ref_M = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_ref_tot-STDFAC*IQUV_smartg_std_ref_tot, print_res=False)

    # Compute the delta_m values from the smartg test results
    delta_m = compute_deltam(obs=IQUV_mystic_tot, mod=IQUV_smartg_tot, print_res=False)
    logger.info(f'A2 - I={delta_m[0]:.3f}; Q={delta_m[1]:.3f}; U={delta_m[2]:.3f}; V={delta_m[3]:.3f} - calculated delta_m')

    # Check if the the test is ok by comparing smartg ref and smartg test
    IQUV_name = ['I', 'Q', 'U', 'V']
    for istk, stk in enumerate(IQUV_name):
        maxVal = max(delta_m_ref[istk], delta_m_ref_P[istk], delta_m_ref_M[istk])
        assert not ( delta_m[istk] > maxVal ), f'Problem with {stk} values, get {delta_m[istk]:.5f}. {stk} must be < to {maxVal:.5f}'

def test_A5_pp(request, S1DF):
    print("=== Test A5 principal plane:")
    # === Atmosphere profil
    z = np.array([1., 0.])
    mol_sca = np.array([0., 0.])[None,:]
    mol_abs= np.array([0., 0.])[None,:]
    cld_tau_ext = np.full_like(mol_sca, 5., dtype=np.float32)
    cld_tau_ext[:,0] = 0. # dtau TOA equal to 0
    cld_ssa = np.full_like(mol_sca, 0.999979, dtype=np.float32)
    prof_aer = (cld_tau_ext, cld_ssa)
    NTH = 18001  # The water cloud has a phase function with a non-negligible peak, then a sufficiently fine resolution is required.
    file_cld_phase = DIR_AUXDATA / 'IPRT' / 'phaseA' / 'opt_prop' / 'watercloud.mie.cdf'
    cld_phase = read_cld_nth_cte(filename=file_cld_phase, nb_theta=NTH)
    pha_atm, ipha_atm = calc_iphase(cld_phase, np.array([800.]), z)
    lpha_lut = []
    for i in range (0, pha_atm.shape[0]): lpha_lut.append(LUT(pha_atm[i,:,:], axes=[None, np.linspace(0, 180, NTH)], names=['stk', 'theta_atm'])) 
    atm = AtmAFGL('afglt', grid=z, prof_ray=mol_sca, prof_abs=mol_abs, prof_aer=prof_aer, prof_phases=(ipha_atm, lpha_lut))
    pro = atm.calc(800., phase=False)
    surf  = None

    # === Illumination conditions
    SZA = 50.
    SAA = 0.
    PHI_0 = 180.-SAA # SMART-G anticlockwise converted to be consistent with MYSTIC

    VZAMIN = 100.
    VZAMAX = 180.
    VZAINC = 1.
    VZA = np.arange(VZAMIN, VZAMAX+VZAINC, VZAINC)

    VAA = np.array([0., 180.])

    # SMART-G Forward TH and PHI using local estimate (anticlockwise) conversion with VZA and VAA MYSTIC (clockwise)
    TH  = 180.-VZA
    PHI = -VAA
    TH[TH==0] = 1e-6 # avoid problem due to special case of 0
    le     = {'th_deg':TH, 'phi_deg':PHI}#, 'zip':True}

    # === Simulation
    mA5F_pp = S1DF.run(THVDEG=SZA, PHVDEG=PHI_0, wl=800., NBPHOTONS=1e7, NBLOOP=1e6, NF=NTH,atm=pro, OUTPUT_LAYERS=int(7),
                       le=le, surf=surf, XBLOCK = 64, XGRID = 1024, BEER=1, DEPO=0.03, stdev=True, SEED=SEED)
    
    with TemporaryDirectory() as tmpdir:
        # === Convert smartg output to iprt ascii output format 
        tmp_file_a5_pp =  Path(tmpdir)/'a5_pp.dat'
        convert_SGout_to_IPRTout(lm=[mA5F_pp, mA5F_pp], lU_sign=[-1, -1], case_name="A5", ldepol=[0.03, 0.03], lalt=[0., 1.], lSZA=[50., 50.], lSAA=[0., 0.], lVZA=[VZA, VZA],
                                lVAA=[VAA, VAA], file_name=tmp_file_a5_pp, output_layer=['_down (0+)', '_up (TOA)'])
        
        # === Plot and comparison with MYSTIC (to save in the report)
        smartg_a5_pp = pd.read_csv(tmp_file_a5_pp, header=None, sep=r'\s+', dtype=float, comment="#").values
    mystic_a5_pp = pd.read_csv(DIR_AUXDATA / 'IPRT' / 'phaseA' / 'mystic_res' / 'iprt_case_a5_pp_mystic.dat',
                               header=None, sep=r'\s+', dtype=float, comment="#").values
    VZAn = np.sort(np.concatenate((VZA-180, 180-VZA)))
    NVZA = len(VZAn)

    # Reflectance
    IQUVS_with_std = seclect_iprt_IQUV(smartg_a5_pp, 1., change_U_sign=False, inv_thetas=True, stdev=True)
    IQUVM_with_std = seclect_iprt_IQUV(mystic_a5_pp, 1., change_U_sign=False, inv_thetas=True, I_index=5, va_index=3, phi_index=4, z_index=0, stdev=True)

    # MYSTIC IQUV and stdev IQUV
    IQUVM_pp = np.zeros((4,NVZA), dtype=np.float32)
    IQUVstdM_pp = np.zeros((4,NVZA), dtype=np.float32)
    IQUVS_pp = np.zeros((4,NVZA), dtype=np.float32)
    IQUVstdS_pp = np.zeros((4,NVZA), dtype=np.float32)
    for i in range (0, 4):
        IQUVM_pp[i,:]=np.concatenate((IQUVM_with_std[i][:,1], IQUVM_with_std[i][::-1,0]))
        IQUVstdM_pp[i,:]=np.concatenate((IQUVM_with_std[i+4][:,1], IQUVM_with_std[i+4][::-1,0]))
        IQUVS_pp[i,:]=np.concatenate((IQUVS_with_std[i][:,1], IQUVS_with_std[i][::-1,0]))
        IQUVstdS_pp[i,:]=np.concatenate((IQUVS_with_std[i+4][:,1], IQUVS_with_std[i+4][::-1,0]))

    IQUVS_pp_tot = IQUVS_pp.copy()
    IQUVM_pp_tot = IQUVM_pp.copy()

    IQUVyMin = [0., -2e-2, -1.2e-4, -1e-5]
    IQUVyMax = [2.5e-1, 1.5e-2, 6e-5, 1e-5]
    plot_iprt_radiances(IQUV_obs=IQUVM_pp, IQUV_mod=IQUVS_pp, IQUVstd_obs=IQUVstdM_pp, IQUVstd_mod=IQUVstdS_pp, xaxis=VZAn,
                        xlabel= 'VZA [deg]', IQUVyMin=IQUVyMin, IQUVyMax=IQUVyMax, title='reflectance  MYSTIC-red SMARTG-blue')
    conftest.savefig(request, bbox_inches='tight')

    # Transmittance
    IQUVS_with_std = seclect_iprt_IQUV(smartg_a5_pp, 0., change_U_sign=False, inv_thetas=True, stdev=True)
    IQUVM_with_std = seclect_iprt_IQUV(mystic_a5_pp, 0., change_U_sign=False, inv_thetas=True, I_index=5, va_index=3, phi_index=4, z_index=0, stdev=True)

    # MYSTIC IQUV and stdev IQUV
    for i in range (0, 4):
        IQUVM_pp[i,:]=np.concatenate((IQUVM_with_std[i][:,1], IQUVM_with_std[i][::-1,0]))
        IQUVstdM_pp[i,:]=np.concatenate((IQUVM_with_std[i+4][:,1], IQUVM_with_std[i+4][::-1,0]))
        IQUVS_pp[i,:]=np.concatenate((IQUVS_with_std[i][:,1], IQUVS_with_std[i][::-1,0]))
        IQUVstdS_pp[i,:]=np.concatenate((IQUVS_with_std[i+4][:,1], IQUVS_with_std[i+4][::-1,0]))

    IQUVyMin = [0., -3e-3, -1.5e-4, -2e-5]
    IQUVyMax = [3.5, 4e-3, 2e-4, 3e-5]

    plot_iprt_radiances(IQUV_obs=IQUVM_pp, IQUV_mod=IQUVS_pp, IQUVstd_obs=IQUVstdM_pp, IQUVstd_mod=IQUVstdS_pp, xaxis=VZAn,
                        xlabel= 'VZA [deg]', IQUVyMin=IQUVyMin, IQUVyMax=IQUVyMax, title='transmittance  MYSTIC-red SMARTG-blue')
    conftest.savefig(request, bbox_inches='tight')

    IQUVS_pp_tot = np.concatenate((IQUVS_pp_tot, IQUVS_pp), axis=1)
    IQUVM_pp_tot = np.concatenate((IQUVM_pp_tot, IQUVM_pp), axis=1)

    # === Compute the delta_m values and analyse them with the previous saved validated ones
    # SMARTG ref results
    smartg_a5_pp_ref = pd.read_csv(DIR_AUXDATA / 'IPRT'/ 'phaseA' / 'smartg_ref_res' / 'iprt_output_format' / 'iprt_case_a5_smartg_pp_ref.dat',
                                   header=None, sep=r'\s+', dtype=float, comment="#").values
    IQUVS_with_std_ref = seclect_iprt_IQUV(smartg_a5_pp_ref, 1., change_U_sign=False, inv_thetas=True, stdev=True)
    IQUVS_pp_ref = np.zeros((4,NVZA), dtype=np.float32)
    IQUVS_pp_std_ref = np.zeros((4,NVZA), dtype=np.float32)
    for i in range (0, 4): 
        IQUVS_pp_ref[i,:]     = np.concatenate((IQUVS_with_std_ref[i][:,1], IQUVS_with_std_ref[i][::-1,0]))
        IQUVS_pp_std_ref[i,:] = np.concatenate((IQUVS_with_std_ref[i+4][:,1], IQUVS_with_std_ref[i+4][::-1,0]))
    IQUVS_pp_ref_tot = IQUVS_pp_ref.copy()
    IQUVS_pp_std_ref_tot = IQUVS_pp_std_ref.copy()
    IQUVS_with_std_ref = seclect_iprt_IQUV(smartg_a5_pp_ref, 0., change_U_sign=False, inv_thetas=True, stdev=True)
    for i in range (0, 4):
        IQUVS_pp_ref[i,:]     = np.concatenate((IQUVS_with_std_ref[i][:,1], IQUVS_with_std_ref[i][::-1,0]))
        IQUVS_pp_std_ref[i,:] = np.concatenate((IQUVS_with_std_ref[i+4][:,1], IQUVS_with_std_ref[i+4][::-1,0]))
    IQUVS_pp_ref_tot = np.concatenate((IQUVS_pp_ref_tot, IQUVS_pp_ref), axis=1)
    IQUVS_pp_std_ref_tot = np.concatenate((IQUVS_pp_std_ref_tot, IQUVS_pp_std_ref), axis=1)

    # Compute the delta_m values from the ref smartg results
    delta_m_ref = compute_deltam(obs=IQUVM_pp_tot, mod=IQUVS_pp_ref_tot, print_res=False)
    logger.info(f'A5_pp - I={delta_m_ref[0]:.3f}; Q={delta_m_ref[1]:.3f}; U={delta_m_ref[2]:.3f}; V={delta_m_ref[3]:.3f} - ref delta_m:')

    # Compute the delta_m values from the ref smartg results +- err
    delta_m_ref_P = compute_deltam(obs=IQUVM_pp_tot, mod=IQUVS_pp_ref_tot+STDFAC*IQUVS_pp_std_ref_tot, print_res=False)
    delta_m_ref_M = compute_deltam(obs=IQUVM_pp_tot, mod=IQUVS_pp_ref_tot-STDFAC*IQUVS_pp_std_ref_tot, print_res=False)

    # Compute the delta_m values from the smartg test results
    delta_m = compute_deltam(obs=IQUVM_pp_tot, mod=IQUVS_pp_tot, print_res=False)
    logger.info(f'A5_pp - I={delta_m[0]:.3f}; Q={delta_m[1]:.3f}; U={delta_m[2]:.3f}; V={delta_m[3]:.3f} - calculated delta_m:')

     # Check if the the test is ok by comparing smartg ref and smartg test
    IQUV_name = ['I', 'Q', 'U', 'V']
    for istk, stk in enumerate(IQUV_name):
        maxVal = max(delta_m_ref[istk], delta_m_ref_P[istk], delta_m_ref_M[istk])
        assert not ( delta_m[istk] > maxVal ), f'Problem with {stk} values, get {delta_m[istk]:.5f}. {stk} must be < to {maxVal:.5f}'


def test_A5_al(request, S1DF):
    print("=== Test A5 almucantar:")
    # === Atmosphere profil
    z = np.array([1., 0.])
    mol_sca = np.array([0., 0.])[None,:]
    mol_abs= np.array([0., 0.])[None,:]
    cld_tau_ext = np.full_like(mol_sca, 5., dtype=np.float32)
    cld_tau_ext[:,0] = 0. # dtau TOA equal to 0
    cld_ssa = np.full_like(mol_sca, 0.999979, dtype=np.float32)
    prof_aer = (cld_tau_ext, cld_ssa)
    NTH = 18001  # The water cloud has a phase function with a non-negligible peak, then a sufficiently fine resolution is required.
    file_cld_phase = DIR_AUXDATA / 'IPRT' / 'phaseA' / 'opt_prop' / 'watercloud.mie.cdf'
    cld_phase = read_cld_nth_cte(filename=file_cld_phase, nb_theta=NTH)
    pha_atm, ipha_atm = calc_iphase(cld_phase, np.array([800.]), z)
    lpha_lut = []
    for i in range (0, pha_atm.shape[0]): lpha_lut.append(LUT(pha_atm[i,:,:], axes=[None, np.linspace(0, 180, NTH)], names=['stk', 'theta_atm'])) 
    atm = AtmAFGL('afglt', grid=z, prof_ray=mol_sca, prof_abs=mol_abs, prof_aer=prof_aer, prof_phases=(ipha_atm, lpha_lut))
    pro = atm.calc(800., phase=False)
    surf  = None

    # === Illumination conditions
    SZA = 50.
    SAA = 0.
    PHI_0 = 180.-SAA # SMART-G anticlockwise converted to be consistent with MYSTIC

    VZA = np.array([130.])

    VAAMIN = 0.
    VAAMAX = 180.
    VAAINC = 1.
    VAA = np.arange(VAAMIN, VAAMAX+VAAINC, VAAINC)

    # SMART-G Forward TH and PHI using local estimate (anticlockwise) conversion with VZA and VAA MYSTIC (clockwise)
    TH  = 180.-VZA
    PHI = -VAA
    TH[TH==0] = 1e-6 # avoid problem due to special case of 0
    le     = {'th_deg':TH, 'phi_deg':PHI}

    # === Simulation
    mA5F_al = S1DF.run(THVDEG=SZA, PHVDEG=PHI_0, wl=800., NBPHOTONS=1e7, NBLOOP=1e6, NF=NTH,atm=pro, OUTPUT_LAYERS=int(7),
                       le=le, surf=surf, XBLOCK = 64, XGRID = 1024, BEER=1, DEPO=0.03, stdev=True, SEED=SEED)
    
    with TemporaryDirectory() as tmpdir:
        # === Convert smartg output to iprt ascii output format 
        tmp_file_a5_al = Path(tmpdir)/'a5_al.dat'
        convert_SGout_to_IPRTout(lm=[mA5F_al, mA5F_al], lU_sign=[-1, -1], case_name="A5", ldepol=[0.03, 0.03], lalt=[0., 1.], lSZA=[50., 50.], lSAA=[0., 0.], lVZA=[VZA, VZA],
                                lVAA=[VAA, VAA], file_name=tmp_file_a5_al, output_layer=['_down (0+)', '_up (TOA)'])
        
        # === Plot and comparison with MYSTIC (to save in the report)
        smartg_a5_al = pd.read_csv(tmp_file_a5_al, header=None, sep=r'\s+', dtype=float, comment="#").values
    mystic_a5_al = pd.read_csv(DIR_AUXDATA / 'IPRT' / 'phaseA' / 'mystic_res' / 'iprt_case_a5_al_mystic.dat',
                               header=None, sep=r'\s+', dtype=float, comment="#").values
    VAAn = VAA
    NVAA = len(VAAn)

    # Reflectance
    IQUVS_with_std = seclect_iprt_IQUV(smartg_a5_al, 1., change_U_sign=False, inv_thetas=True, stdev=True)
    IQUVM_with_std = seclect_iprt_IQUV(mystic_a5_al, 1., change_U_sign=False, inv_thetas=True, I_index=5, va_index=3, phi_index=4, z_index=0, stdev=True)

    # MYSTIC IQUV and stdev IQUV
    IQUVM_al = np.zeros((4,NVAA), dtype=np.float32)
    IQUVstdM_al = np.zeros((4,NVAA), dtype=np.float32)
    IQUVS_al = np.zeros((4,NVAA), dtype=np.float32)
    IQUVstdS_al = np.zeros((4,NVAA), dtype=np.float32)
    for i in range (0, 4):
        IQUVM_al[i,:]=IQUVM_with_std[i][0,:]
        IQUVstdM_al[i,:]=IQUVM_with_std[i+4][0,:]
        IQUVS_al[i,:]=IQUVS_with_std[i][0,:]
        IQUVstdS_al[i,:]=IQUVS_with_std[i+4][0,:]

    IQUVS_al_tot = IQUVS_al.copy()
    IQUVM_al_tot = IQUVM_al.copy()

    IQUVyMin = [6e-2, -1e-2, -2e-3, -5e-5]
    IQUVyMax = [1.2e-1, 2e-2, 1.2e-2, 2e-5]
    plot_iprt_radiances(IQUV_obs=IQUVM_al, IQUV_mod=IQUVS_al, IQUVstd_obs=IQUVstdM_al, IQUVstd_mod=IQUVstdS_al, xaxis=VAAn,
                        xlabel= 'VZA [deg]', IQUVyMin=IQUVyMin, IQUVyMax=IQUVyMax, title='reflectance  MYSTIC-red SMARTG-blue')
    conftest.savefig(request, bbox_inches='tight')

    # Transmittance
    IQUVS_with_std = seclect_iprt_IQUV(smartg_a5_al, 0., change_U_sign=False, inv_thetas=True, stdev=True)
    IQUVM_with_std = seclect_iprt_IQUV(mystic_a5_al, 0., change_U_sign=False, inv_thetas=True, I_index=5, va_index=3, phi_index=4, z_index=0, stdev=True)

    # MYSTIC IQUV and stdev IQUV
    for i in range (0, 4):
        IQUVM_al[i,:]=IQUVM_with_std[i][0,:]
        IQUVstdM_al[i,:]=IQUVM_with_std[i+4][0,:]
        IQUVS_al[i,:]=IQUVS_with_std[i][0,:]
        IQUVstdS_al[i,:]=IQUVS_with_std[i+4][0,:]

    IQUVyMin = [0., -3.5e-3, -3e-3, -1.5e-5]
    IQUVyMax = [3.5, 5e-4, 5e-4, 2.5e-5]

    plot_iprt_radiances(IQUV_obs=IQUVM_al, IQUV_mod=IQUVS_al, IQUVstd_obs=IQUVstdM_al, IQUVstd_mod=IQUVstdS_al, xaxis=VAAn,
                        xlabel= 'VZA [deg]', IQUVyMin=IQUVyMin, IQUVyMax=IQUVyMax, title='transmittance  MYSTIC-red SMARTG-blue')
    conftest.savefig(request, bbox_inches='tight')

    IQUVS_al_tot = np.concatenate((IQUVS_al_tot, IQUVS_al), axis=1)
    IQUVM_al_tot = np.concatenate((IQUVM_al_tot, IQUVM_al), axis=1)

    # === Compute the delta_m values and analyse them with the previous saved validated ones
    # SMARTG ref results
    smartg_a5_al_ref = pd.read_csv(DIR_AUXDATA / 'IPRT' / 'phaseA' / 'smartg_ref_res' / 'iprt_output_format' / 'iprt_case_a5_smartg_al_ref.dat',
                                   header=None, sep=r'\s+', dtype=float, comment="#").values
    IQUVS_with_std_ref = seclect_iprt_IQUV(smartg_a5_al_ref, 1., change_U_sign=False, inv_thetas=True, stdev=True)
    IQUVS_al_ref = np.zeros((4,NVAA), dtype=np.float32)
    IQUVS_al_std_ref = np.zeros((4,NVAA), dtype=np.float32)
    for i in range (0, 4): 
        IQUVS_al_ref[i,:]     = IQUVS_with_std_ref[i][0,:]
        IQUVS_al_std_ref[i,:] = IQUVS_with_std_ref[i+4][0,:]
    IQUVS_al_ref_tot = IQUVS_al_ref.copy()
    IQUVS_al_std_ref_tot = IQUVS_al_std_ref.copy()
    IQUVS_with_std_ref = seclect_iprt_IQUV(smartg_a5_al_ref, 0., change_U_sign=False, inv_thetas=True, stdev=True)
    for i in range (0, 4):
        IQUVS_al_ref[i,:]     = IQUVS_with_std_ref[i][0,:]
        IQUVS_al_std_ref[i,:] = IQUVS_with_std_ref[i+4][0,:]
    IQUVS_al_ref_tot = np.concatenate((IQUVS_al_ref_tot, IQUVS_al_ref), axis=1)
    IQUVS_al_std_ref_tot = np.concatenate((IQUVS_al_std_ref_tot, IQUVS_al_std_ref), axis=1)

    # Compute the delta_m values from the ref smartg results
    delta_m_ref = compute_deltam(obs=IQUVM_al_tot, mod=IQUVS_al_ref_tot, print_res=False)
    logger.info(f'A5_al - I={delta_m_ref[0]:.3f}; Q={delta_m_ref[1]:.3f}; U={delta_m_ref[2]:.3f}; V={delta_m_ref[3]:.3f} - ref delta_m:')

    # Compute the delta_m values from the ref smartg results +- err
    delta_m_ref_P = compute_deltam(obs=IQUVM_al_tot, mod=IQUVS_al_ref_tot+STDFAC*IQUVS_al_std_ref_tot, print_res=False)
    delta_m_ref_M = compute_deltam(obs=IQUVM_al_tot, mod=IQUVS_al_ref_tot-STDFAC*IQUVS_al_std_ref_tot, print_res=False)

    # Compute the delta_m values from the smartg test results
    delta_m = compute_deltam(obs=IQUVM_al_tot, mod=IQUVS_al_tot, print_res=False)
    logger.info(f'A5_al - I={delta_m[0]:.3f}; Q={delta_m[1]:.3f}; U={delta_m[2]:.3f}; V={delta_m[3]:.3f} - calculated delta_m:')

     # Check if the the test is ok by comparing smartg ref and smartg test
    IQUV_name = ['I', 'Q', 'U', 'V']
    for istk, stk in enumerate(IQUV_name):
        maxVal = max(delta_m_ref[istk], delta_m_ref_P[istk], delta_m_ref_M[istk])
        assert not ( delta_m[istk] > maxVal ), f'Problem with {stk} values, get {delta_m[istk]:.5f}. {stk} must be < to {maxVal:.5f}'


import papermill as pm

from pathlib import Path
ROOTPATH =  Path(__file__).resolve().parent.parent.parent


def test_demo_notebook():
    """
    Execute the demo notebook
    """
    print("\nTesting demo_notebook.ipynb...")
    nb_path = ROOTPATH / 'smartg' / 'notebooks' / 'demo_notebook.ipynb'
    nb_output_path = ROOTPATH / 'smartg' / 'tests' / 'logs' / 'demo_notebook_log.ipynb'
    pm.execute_notebook(nb_path, nb_output_path, cwd=ROOTPATH)

def test_demo_notebook_objects():
    """
    Execute the demo notebook objects
    """
    print("\nTesting demo_notebook_objects.ipynb...")
    nb_path = ROOTPATH / 'smartg' / 'notebooks' / 'demo_notebook_objects.ipynb'
    nb_output_path = ROOTPATH / 'smartg' / 'tests' / 'logs' / 'demo_notebook_objects_log.ipynb'
    pm.execute_notebook(nb_path, nb_output_path, cwd=ROOTPATH)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from urllib.request import urlretrieve
import zipfile
import tarfile

dir_root = Path(__file__).resolve().parent.parent

# auxdata source: HYGEOS
AER_URL = "https://docs.hygeos.com/s/8PnKXFXQbmYyTte/download"
ACS_URL = "https://docs.hygeos.com/s/HwotAHPstdCCKcJ/download"
ATM_URL = "https://docs.hygeos.com/s/z6MRf9g66WmWeBA/download"
STP_URL = "https://docs.hygeos.com/s/NW42DNPtKw3NNW7/download"
VALID_URL = "https://docs.hygeos.com/s/6EPBqwebn94NYPq/download"
WATER_URL = "https://docs.hygeos.com/s/3NKP5tMsHKnNRpt/download"
KDIS_URL = "https://docs.hygeos.com/s/CHTFFgHe6to39CR/download"
CLOUD_URL = "https://docs.hygeos.com/s/agDWDy998j64SHf/download"

# some data (mystic res and opt_prop) are taken from: 
# https://www.meteo.physik.uni-muenchen.de/~iprt/doku.php?id=intercomparisons:intercomparisons
IPRT_URL = "https://docs.hygeos.com/s/i4QaxtpjSfjwtNk/download"

# reptran source: http://www.libradtran.org
# the libradtran url is better to be sure to get the last versions of reptran look-up tables
REPTRAN_URL = "http://www.meteo.physik.uni-muenchen.de/~libradtran/lib/exe/fetch.php?media=download:reptran_2017_all.tar.gz"
# since the above url is not always stable, we provide an alternative url (but can be outdated!!)
REPTRAN_URL_HYG = "https://docs.hygeos.com/s/jHKMcZZmkf6xy7D/download"

AUXDATA_DICT = {
    "aer": AER_URL,
    "acs": ACS_URL,
    "atm": ATM_URL,
    "STP": STP_URL,
    "valid": VALID_URL,
    "water": WATER_URL,
    "kdis": KDIS_URL,
    "cld": CLOUD_URL,
    "IPRT": IPRT_URL,
    "reptran": REPTRAN_URL,
}


def safe_download(url, outfile):

    def reporthook(count, block_size, total_size):
        if total_size > 0:
            percent = int(count * block_size * 100 / total_size) if total_size > 0 else 0
            print(f"\rDownloading {outfile}: {percent}%", end="")
        else:
            downloaded = count * block_size
            print(f"\rDownloaded {downloaded/1024/1024:.1f} MB...", end="")

    print(f"Downloading {url} → {outfile}")
    urlretrieve(url, outfile, reporthook)
    print("\nDownload complete.")


def extract_zip(zfile, dest):
    with zipfile.ZipFile(zfile, 'r') as z:
        print(f"Extracting ZIP {zfile} → {dest}")
        for name in z.namelist():
            print("  extracting:", name)
        z.extractall(dest)


def extract_tar(tfile, dest):
    with tarfile.open(tfile, "r:gz") as tar:
        print(f"Extracting TAR {tfile} → {dest}")
        for member in tar.getmembers():
            print("  extracting:", member.name)
        tar.extractall(dest)


def download(savepath, data_type="all"):
    """
    Dowload the SMART-G auxiliary data

    Parameters
    ----------
    savepath : str
        The path where the data will be saved
    data_type : str, optional
        The type of data to download, can be: "all", "aer", "acs", "atm", "STP", "valid", 
        "water", "kdis", "cld", "IPRT". Default is "all". Definitions:

        * all -> all the available data
        * aer -> aerosols data
        * acs -> absorption cross section coefficients data
        * atm -> atmosphere profils
        * STP -> STP (Solar Power Tower) files with heliostat positions
        * valid -> validation files
        * water -> water files needed for some simulations including the ocean
        * kdis -> k-distribution
        * cld -> cloud data
        * IPRT -> some data from IPRT (International working group on Polarized Radiative Transfer)

    Examples
    --------
    >>> from pathlib import Path
    >>> from smartg.auxdata import download
    >>> savepath = Path("/dir/where/to/save/data")
    >>> download(savepath, data_type="all")
    """

    list_kind = ["all"] + list(AUXDATA_DICT.keys())

    if data_type not in list_kind:
        raise ValueError("Invalid value for 'kind'. Must be one of: " + ", ".join(list_kind))
    
    savepath = Path(savepath)
    savepath.mkdir(parents=True, exist_ok=True)

    if data_type == "all": names = list(AUXDATA_DICT.keys())
    else                 : names = [data_type]

    for name in names:
        try:
            print(f"Trying to download {name} auxiliary data in {savepath}...\n")

            if name == "reptran":
                out = savepath / f"{name}.tar.gz"
                safe_download(AUXDATA_DICT[name], out)
                extract_tar(out, savepath)
                Path(out).unlink(missing_ok=True)

            else:
                out = savepath / f"{name}.zip"
                safe_download(AUXDATA_DICT[name] + "/" + name + ".zip", out)
                extract_zip(out, savepath)
                Path(out).unlink(missing_ok=True)

            print(f"{name} auxiliary data downloaded and extracted successfully. ✅\n")

        except Exception as e1:
            print(f"Error during download and/or extraction: {e1}. ❌\n")

            if name == "reptran":
                print("Another url is available for reptran, trying again...\n")
                try:
                    out = savepath / f"{name}.zip"
                    safe_download(f"{REPTRAN_URL_HYG}/{name}.zip", out)
                    extract_zip(out, savepath)
                    Path(out).unlink(missing_ok=True)
                    print(f"{name} auxiliary data downloaded and extracted successfully. ✅\n")
                except Exception as e2:
                    print(f"Error during download and/or extraction: {e2}. ❌\n")


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division
from numpy import sin, cos, pi
import numpy as np
#from luts.luts import LUT

def fournierForand(ang, n, mu):
    '''
    Fournier-Forand phase function
    Parameters:
        ang: angle in radians
        n: index of refraction of the particles
        mu: slope parameter of the hyperbolic dustribution
    Normalized to 4pi
    See http://www.oceanopticsbook.info/view/scattering/the_fournierforand_phase_function
    '''
    v = (3-mu)/2
    delta = 4/( 3*(n-1)*(n-1) )*sin(ang/2)*sin(ang/2)
    delta180 = 4/( 3*(n-1)*(n-1) )*sin(pi/2)*sin(pi/2)

    with np.errstate(divide='ignore', invalid='ignore'):
        res = 1/( 4*pi*(1-delta)*(1-delta)*(delta**v) )*( v*(1-delta) - (1-(delta**v)) + ( delta*(1-(delta**v)) - v*(1-delta) )*1/(sin(ang/2)*sin(ang/2)) ) + (1-(delta180**v))/(16*pi*(delta180-1)*(delta180**v)) * (3*cos(ang)*cos(ang) - 1)
    res *= 4*pi

    return res

def fournierForandB(n, mu):
    '''
    backscatter fraction of the Fournier-Forand phase function
    '''
    d90 = 4./(3.*(n-1.)**2)*(sin(pi/4.)**2)
    v = (3.-mu)/2.
    B   = 1 - (1 - d90**(v+1) - 0.5*(1-d90**v))/((1-d90)*d90**v)
    return B

def henyeyGreenstein(angle, g):
    '''
    Henyey-Greenstein phase function
    Parameters:
        angle: angle in radians
        g: asymmetry coefficient
           (0: isotropic ; 1: highly peaked)
    Normalized to 4pi
    See http://www.oceanopticsbook.info/view/scattering/the_henyeygreenstein_phase_function
    '''
    return (1 - g*g)/((1 + g*g - 2*g*cos(angle))**1.5)

def integ_phase(ang, pha):
    '''
    Integrate pha(ang)*sin(ang) along the last axis
    ang in radians
    pha: phase function, dim [...,ang]
    '''
    assert not np.isnan(pha).any()

    dtheta = np.diff(ang)
    pm1 = pha[...,:-1]
    pm2 = pha[...,1:]
    sin1 = np.sin(ang[:-1])
    sin2 = np.sin(ang[1:])

    return np.sum(dtheta*((sin1*pm1+sin2*pm2)/3. + (sin1*pm2+sin2*pm1)/6.), axis=-1)

def calc_iphase(phase, wav_full, z_full, old_method=False):
    '''
    calculate phase function indices
    phase is a LUT of shape [wav, z, stk, theta]

    returns (pha, ipha) where:
        * pha is an array reshaped from phase to [wav*z, stk, theta]
        * ipha is an array of phase function indices (starting from 0)
          in the full array [wav_full, z_full]
    '''
    wav = phase.axes[0]
    altitude = phase.axes[1]

    nwav, nz, nstk, ntheta = phase.shape
    pha = phase.data.reshape(nwav*nz, nstk, ntheta)

    ipha_w = np.array([np.abs(wav - x).argmin() for x in wav_full], dtype='int32')
    if old_method:
        ipha_a = np.array([np.abs(altitude - x).argmin() for x in z_full], dtype='int32')
    else:
        ipha_a = get_ipha_a(z_full=z_full, z_pf=altitude, phase=phase)
    ipha = ipha_a[None,:] + ipha_w[:,None]*len(altitude)

    return (pha, ipha)


def get_ipha_a(z_full, z_pf, phase=None):
    # Particular case with only 1 phase matrix for the whole z column
    if len(z_pf) == 1:
        ida = np.zeros_like(z_full, dtype=np.int32)
        return ida
    
    grid_full = z_full
    grid_pf = z_pf
    size_layers_full = np.concatenate(( np.array([1e6]), np.abs(np.diff(grid_full))))
    size_layers_pf = np.concatenate(( np.array([1e6]), np.abs(np.diff(grid_pf))))

    nz_full = len(grid_full)
    nz_pf = len(grid_pf)

    zmin_print = [-1e8]
    zmax_print = [-1e8]

    ida = np.full(nz_full, -1, dtype=np.int32)
    for i_full in range (0, nz_full):
        idz_full = (nz_full-1)-i_full
        zmin_full = grid_full[idz_full]
        zmax_full = grid_full[idz_full] + size_layers_full[idz_full]

        # First find all the z_pf layers respecting the 2 conditions
        ida_tmp = []
        for i_pf in range(0, nz_pf):
            idz_pf = (nz_pf-1)-i_pf
            zmin_pf = grid_pf[idz_pf]
            zmax_pf = grid_pf[idz_pf] + size_layers_pf[idz_pf]
            cond_1 = zmin_pf < zmax_full
            cond_2 = zmax_pf > zmin_full
            if (cond_1 and cond_2):
                ida_tmp.append(idz_pf)

        n_ida_tmp = len(ida_tmp)
        # if only one pf layer respect the conditions take directly its index
        if n_ida_tmp == 1 :
            ida[idz_full] = ida_tmp[0]
        # if more than one, we have to look which pf layer fill best the layer of z_full 
        elif( n_ida_tmp > 1):
            pfs_weight = np.zeros(n_ida_tmp)
            for k in range (0, n_ida_tmp):
                zmin_pf_k = grid_pf[ida_tmp[k]]
                zmax_pf_k = grid_pf[ida_tmp[k]] + size_layers_pf[ida_tmp[k]]
                pf_full_min = max(zmin_pf_k, zmin_full)
                pf_full_max = min(zmax_pf_k, zmax_full)
                if ( (phase == None) or (np.sum(phase[0,ida_tmp[k],0,:]) > 0.) ):
                    pfs_weight[k] = pf_full_max - pf_full_min
                else:
                    if (zmax_pf_k < 1e6) and ( (zmin_pf_k not in zmin_print) and (zmax_pf_k not in zmax_print)):
                        print("Warning: null phase matrix between ", zmin_pf_k, " and ", zmax_pf_k,
                              " detected! Please check pfgrid and/or grid (z_atm) values.")
                        zmin_print.append(zmin_pf_k)
                        zmax_print.append(zmax_pf_k)
                    pfs_weight[k] = (pf_full_max - pf_full_min)*1e-6
            ida[idz_full] = ida_tmp[np.argmax(pfs_weight)]
        elif (n_ida_tmp == 0 and np.sum(z_full) < 0.): # Particular case of min z_pf > min z_full in ocean
            ida[idz_full] = int(len(z_pf) - 1)
    return ida



import numpy as np
cimport numpy as np

from pathlib import Path
from clut cimport CLUT
from libc.math cimport exp, M_PI, isnan, log
from warnings import warn



cdef class WaterModel:
    '''
    Base class for water reflectance models
    '''
    cdef init(self, float[:] wav, float sza, float vza, float raa):
        raise Exception('WaterModel.init(...) shall be implemented')

    cdef float[:] calc_rho(self, float[:] x):
        raise Exception('WaterModel.calc_rho(...) shall be implemented')


cdef class ParkRuddick(WaterModel):

    cdef float[:] wav  # wavelength (nm)
    cdef float[:] Rw   # water reflectance (0+)
    cdef float[:] bw   # pure water scattering coefficient
    cdef float[:] aw   # pure water absorption coefficient
    cdef float[:] a_bric    # absorption parameters A and B (Bricaud)
    cdef float[:] b_bric    # absorption parameters A and B (Bricaud)
    cdef float[:] a_star    # mineral absorption coefficient

    cdef CLUT BW
    cdef CLUT AW_POPEFRY
    cdef CLUT AW_PALMERW
    cdef CLUT GI_PR  # gi coefficients
    cdef CLUT GII_PR  # pre-interpolated gi coefficients (for one pixel)
    cdef CLUT AB_BRIC
    cdef CLUT RAMAN
    cdef CLUT ASTAR
    cdef float bw500
    cdef float a700
    cdef float mus
    cdef int Nwav
    cdef int debug
    cdef int alt_gamma_bb
    cdef int min_abs
    cdef float[:] atot, bbtot, btot, aphy, aCDM, aNAP
    cdef float gamma, SPM
    cdef object out_type

    cdef int[:] index  # multi-purpose vector

    def __init__(self, directory, alt_gamma_bb=False, min_abs=False, debug=False):

        self.Rw = None
        self.index = np.zeros(2, dtype='int32')
        self.debug = debug
        self.alt_gamma_bb = alt_gamma_bb
        self.min_abs = min_abs  # activate mineral absorption
        directory = Path(directory)
        self.read_iop(directory)
        self.read_gi(directory)

    def read_iop(self, directory):

        #
        # read water scattering coefficient
        #
        directory = Path(directory)
        data_bw = np.genfromtxt(directory / 'morel_buiteveld_bsw.txt', skip_header=1)
        self.BW = CLUT(data_bw[:,1], axes=[data_bw[:,0]])
        assert data_bw[-1,0] == 500.
        self.bw500 = data_bw[-1,1]

        #
        # read pure water absorption coefficient
        #
        # Pope&Fry
        data = np.genfromtxt(directory / 'pope97.dat', skip_header=6)
        data[:,1] *= 100 #  convert from cm-1 to m-1
        self.AW_POPEFRY = CLUT(data[:,1], axes=[data[:,0]])
        # Palmer&Williams
        data = np.genfromtxt(directory / 'palmer74.dat', skip_header=5)
        data[:,1] *= 100 #  convert from cm-1 to m-1
        self.AW_PALMERW = CLUT(data[:,1], axes=[data[:,0]])

        #
        # read phytoplankton absorption
        #
        ap_bricaud = np.loadtxt(directory / 'aph_bricaud_1995.txt', delimiter=',')
        lambda_bric95 = ap_bricaud[:, 0]
        self.AB_BRIC = CLUT(ap_bricaud[:,1:], axes=[lambda_bric95, None])
        assert lambda_bric95[-1] == 700
        self.a700 = ap_bricaud[-1,1]

        #
        # read mineral absorption
        #
        astar_ = np.genfromtxt(directory / 'astarmin_average_2015_SLSTR.txt', comments='%')
        self.ASTAR = CLUT(astar_[:-1,1],
                          axes=[astar_[:-1,0]])

        #
        # read Raman correction
        #
        raman = np.genfromtxt(directory / 'raman_westberry13.txt', comments='#')
        # (wl, chl)
        self.RAMAN = CLUT(raman[:,1:], axes=[raman[:,0],
                [0.01,0.02,0.03,0.04,0.07,0.1,0.2,0.3,0.5,0.7,1.,2.,5.]])

    def read_gi(self, directory):
        '''
        read gi coefficients
        '''
        directory = Path(directory)
        fp = open(directory / 'AboveRrs_gCoef_w5.dat')

        # read first lines
        fp.readline() # first line useless
        line = fp.readline()
        gb = list(map(float, line[:line.find('=')].split()))
        line = fp.readline()
        th0 = list(map(float, line[:line.find('=')].split()))
        line = fp.readline()
        th = list(map(float, line[:line.find('=')].split()))
        line = fp.readline()
        dphi = list(map(float, line[:line.find('=')].split()))

        # initialize output array
        ngb = len(gb)
        nth0 = len(th0)
        nth = len(th)
        ndphi = len(dphi)
        gi = np.zeros((ngb, 4, nth0, nth, ndphi), dtype='float32') - 999.

        # read blocks
        for (ith0, ith, idphi) in np.ndindex(nth0, nth, ndphi):
            fp.readline()

            # read next 8 lines
            block = ''
            for i in xrange(8): block += fp.readline()

            # format block and fill array slice
            block = np.array(list(map(float, block.split()))).reshape((ngb, 4))
            gi[:,:,ith0, ith, idphi] = block[:,:]

        fp.close()

        self.GI_PR = CLUT(gi, axes=[gb, None, th0, th, dphi])

        # initialize pre-interpolated gi coefficients
        # (empty)
        self.GII_PR = CLUT(np.zeros((ngb, 4)), axes=[gb, None])


    cdef init(self, float[:] wav, float sza, float vza, float raa):
        '''
        initialize the model parameters for current pixel
        '''
        cdef int i, j
        cdef int ret
        self.wav = wav
        self.Nwav = len(wav)
        self.mus = np.cos(sza*np.pi/180.)

        #
        # array initialization (unique)
        #
        if self.Rw is None:
            self.Rw = np.zeros(len(wav), dtype='float32') + np.NaN
            self.bw = np.zeros(len(wav), dtype='float32') + np.NaN
            self.aw = np.zeros(len(wav), dtype='float32') + np.NaN
            self.a_bric = np.zeros(len(wav), dtype='float32') + np.NaN
            self.b_bric = np.zeros(len(wav), dtype='float32') + np.NaN
            self.a_star = np.zeros(len(wav), dtype='float32') + np.NaN

            if self.debug:
                self.atot = np.zeros(len(wav), dtype='float32') + np.NaN
                self.bbtot = np.zeros(len(wav), dtype='float32') + np.NaN
                self.btot = np.zeros(len(wav), dtype='float32') + np.NaN
                self.aCDM = np.zeros(len(wav), dtype='float32') + np.NaN
                self.aNAP = np.zeros(len(wav), dtype='float32') + np.NaN
                self.aphy = np.zeros(len(wav), dtype='float32') + np.NaN
        elif len(wav) != len(self.Rw):
            raise Exception('Invalid length of wav')

        #
        # interpolate scattering coefficient
        #
        for i, w in enumerate(wav):
            ret = self.BW.lookup(0, w)
            if ret > 0:
                self.bw[i] = self.bw500 * (w/500.)**-4.
            elif ret < 0:
                raise Exception('Error in BW lookup')
            else:
                self.bw[i] = self.BW.interp()

        #
        # interpolate absorption coefficients
        #
        for i, w in enumerate(wav):

            ret = self.AW_POPEFRY.lookup(0, w)
            if ret < 0:
                raise Exception('Error in AW_POPEFRY lookup')
            elif ret > 0:
                if self.AW_PALMERW.lookup(0, w) != 0:
                    raise Exception('Error in AW_PALMERW lookup')
                self.aw[i] = self.AW_PALMERW.interp()
            else:
                self.aw[i] = self.AW_POPEFRY.interp()

            ret = self.AB_BRIC.lookup(0, w)
            if (ret < 0) and (w > 399.) and (w <= 400):
                ret = self.AB_BRIC.lookup(0, 400.)

            if ret == 0:
                # success
                self.AB_BRIC.index(1, 0)
                self.a_bric[i] = self.AB_BRIC.interp()
                self.AB_BRIC.index(1, 1)
                self.b_bric[i] = self.AB_BRIC.interp()
            elif ret > 0:  # above axis
                self.a_bric[i] = self.a700 * (w/700.)**(-80.)
                self.b_bric[i] = 0
            else:
                raise Exception('Error in AB_BRIC lookup (lambda={})'.format(w))

            ret = self.ASTAR.lookup(0, w)
            if ret != 0:
                raise Exception('Error on A_STAR lookup (wavelength={})'.format(w))
            else:
                self.a_star[i] = self.ASTAR.interp()


        #
        # empty the pre-interpolated gi coefficients
        #
        for i in range(self.GII_PR.shape[0]):
            for j in range(self.GII_PR.shape[1]):
                self.index[0] = i
                self.index[1] = j
                self.GII_PR.set(np.NaN, self.index)
        #
        # lookup the ths, thv and phi axes
        #
        ret = self.GI_PR.lookup(2, sza)
        if ret != 0: raise Exception('Error in GI_PR sza lookup')
        ret = self.GI_PR.lookup(3, vza)
        if ret != 0: raise Exception('Error in GI_PR vza lookup')
        ret = self.GI_PR.lookup(4, raa)
        if ret != 0: raise Exception('Error in GI_PR raa lookup')


    cdef float[:] calc_rho(self, float[:] x):
        '''
        calculate the reflectance above the water surface at all bands,
        for a parameter vector x
        '''

        # x is [logchl, logfb, logfa] or shorter
        cdef int N = len(x)
        cdef float fa, fb
        cdef float logchl = x[0]
        cdef float chl = 10**logchl
        cdef float bbw, gamma, bp550, bbp550, bb, bbp, bbp650
        cdef float SPM
        cdef float aw, aphy, aCDM, a, aCDM443, S, aNAP
        cdef float lam
        cdef float gammab, omegab, omegapow, gi
        cdef int i, j
        cdef int ret
        cdef int igb
        cdef float rho

        if N >= 2:
            fb = 10**x[1]
        else:
            fb = 1.

        if N >= 3:
            fa = 10**x[2]
        else:
            fa = 1.

        #
        # wavelength-independent parameters
        #

        # phytoplankton scattering
        bp550 = 0.416 * (chl**0.766) * fb
        if (logchl < 2):
            bbp550 = (0.002 + 0.01*(0.5 - 0.25*logchl)) * bp550
        else:
            bbp550 = (0.002 + 0.01*(0.5 - 0.25*2.    )) * bp550

        # spectral dependency
        if self.alt_gamma_bb:
            # alternate version (H. Loisel), improved in turbid waters
            # phytoplankton + other particles
            # dependency function of bbp550
            # - high values: personal comm. Loisel
            # - low bbp values: slope from Antoine LO (2011)
            #   offset adjusted for tangency to Loisel
            if bbp550 < 0.00227:
                gamma = -0.845 * log(bbp550) - 3.13
            else:
                gamma = 0.156 * bbp550**(-0.42)

            if gamma > 4:
                gamma = 4.
        else:
            # phytoplankton + other particles
            # spectral dependency (from Antoine LO (2011))
            gamma = -0.733 * log(chl) + 1.499   # /!\ log is ln
            if gamma < 0: gamma = 0

        bbp650 = bbp550 * (650./550.)**(-gamma)
        SPM = 100.*bbp650  # Neukermans, log10(bbp) = 1.03*log10(SPM)

        # CDM absorption central value
        # from Bricaud et al GBC, 2012 (data from nov 2007)
        aCDM443 = fa * 0.069 * (chl**1.070)   # Bricaud et al, GBC, 2012

        S = 0.00262*(aCDM443**(-0.448))
        if (S > 0.025): S=0.025
        if (S < 0.011): S=0.011

        # wavelength loop
        for i in range(self.Nwav):
            lam = self.wav[i]

            #
            # 1) scattering parameters
            #

            # pure water scattering
            bbw = 0.5*self.bw[i]

            bbp = bbp550 * (lam/550)**(-gamma)

            bb = bbw + bbp


            #
            # 2) absorption parameters
            #

            # pure water absorption
            aw = self.aw[i]

            # phytoplankton absorption
            aphy = self.a_bric[i] * chl**(1-self.b_bric[i])

            aCDM = aCDM443 * exp(-S*(lam - 443))

            # mineral absorption
            if self.min_abs:
                aNAP = self.a_star[i]*SPM
            else:
                aNAP = 0.

            # total absorption
            a = aw + aphy + aCDM + aNAP

            omegab = bb/(a + bb)
            gammab = bbp/bb

            omegapow = 1.
            rho = 0.
            ret = self.GII_PR.lookup(0, gammab)
            # if ret != 0:  # FIXME: test return values
                # raise Exception('GII_PR lookup error: gammab={}'.format(gammab))

            # pre-interpolation
            for igb in range(self.GII_PR._inf[0], self.GII_PR._inf[0]+2):
                if igb >= self.GII_PR.shape[0]:
                    continue
                self.index[0] = igb
                self.index[1] = 0
                if isnan(self.GII_PR.get(self.index)):
                    self.GI_PR.index(0, igb)
                    # NOTE: axes ths, thv and phi have already been lookedup on init()
                    for j in range(4):
                        self.GI_PR.index(1, j)
                        self.index[1] = j
                        self.GII_PR.set(self.GI_PR.interp(), self.index)


            for j in range(4):
                self.GII_PR.index(1, j)

                omegapow *= omegab

                gi = self.GII_PR.interp()

                rho += gi * omegapow

            rho *= M_PI # conversion remote sensing reflectance -> reflectance

            # raman correction
            # TODO: pre-interpolate RAMAN in lam?
            ret = self.RAMAN.lookup(0, lam)
            if (ret < 0) and (lam > 399.) and (lam <= 400):
                ret = self.RAMAN.lookup(0, 400.)
            if ret < 0:
                raise Exception('Error in lookup for RAMAN (lambda={})'.format(lam))
            ret = self.RAMAN.lookup(1, chl)  # clip both ends
            rho *= 1. + (self.RAMAN.interp()*self.mus/0.866)

            self.Rw[i] = rho

            if self.debug:
                self.atot[i] = a
                self.bbtot[i] = bb
                self.btot[i] = self.bw[i] + bp550*(lam/550.)**(-gamma)
                self.aphy[i] = aphy
                self.aCDM[i] = aCDM
                self.aNAP[i] = aNAP
                self.gamma = gamma
                self.SPM = SPM


        return self.Rw

    def calc(self, w, logchl, logfb=0., logfa=0., sza=0., vza=0., raa=0.):
        '''
        water reflectance calculation (python interface)
        w: wavelength (nm) [float, list or array]
        returns above-water reflectance as float, list or array
        '''
        if isinstance(w, np.ndarray):
            self.out_type = np.array
            wav = w.astype('float32')
        elif isinstance(w, list):
            self.out_type = list
            wav = np.array(w, dtype='float32')
        else: # float
            self.out_type = lambda x: x[0]
            wav = np.array([w], dtype='float32')

        self.init(wav.astype('float32'), float(sza), float(vza), float(raa))
        params = np.zeros(3, dtype='float32')
        params[0] = logchl
        params[1] = logfb
        params[2] = logfa
        R = self.calc_rho(params)
        return self.out_type(R)

    def iops(self):
        '''
        returns the IOPs for the previously calculated spectrum
        '''
        assert self.debug

        return {
                'bw':   self.out_type(self.bw),
                'aw':   self.out_type(self.aw),
                'atot': self.out_type(self.atot),
                'bbtot':self.out_type(self.bbtot),
                'btot': self.out_type(self.btot),
                'aphy': self.out_type(self.aphy),
                'aCDM': self.out_type(self.aCDM),
                'aNAP': self.out_type(self.aNAP),
                'gamma': self.gamma,
                'SPM': self.SPM,
                }


cdef class MorelMaritorena(WaterModel):

    cdef float[:] wav  # wavelength (nm)
    cdef int Nwav
    cdef CLUT Kw_tab, bw_tab, Chi_tab, e_tab, simspec
    cdef float[:] Kw_i
    cdef float[:] Chi_i
    cdef float[:] e_i
    cdef float[:] bw_i
    cdef float[:] Rw
    cdef float[:] simspec_i
    cdef float lam_join
    cdef initialized
    cdef object out_type

    cdef int debug
    cdef float[:] bw, atot, bbtot

    def __init__(self, debug=False):

        warn('f/Q is not implemented')
        self.debug = debug

        self.Kw_tab = CLUT(np.array([
                    0.02710, 0.02380, 0.02160, 0.01880, 0.01770, 0.01595, 0.01510, 0.01376, 0.01271, 0.01208,
                    0.01042, 0.00890, 0.00812, 0.00765, 0.00758, 0.00768, 0.00770, 0.00792, 0.00885, 0.00990,
                    0.01148, 0.01182, 0.01188, 0.01211, 0.01251, 0.01320, 0.01444, 0.01526, 0.01660, 0.01885,
                    0.02188, 0.02701, 0.03385, 0.04090, 0.04214, 0.04287, 0.04454, 0.04630, 0.04846, 0.05212,
                    0.05746, 0.06053, 0.06280, 0.06507, 0.07034, 0.07801, 0.09038, 0.11076, 0.13584, 0.16792,
                    0.22310, 0.25838, 0.26506, 0.26843, 0.27612, 0.28400, 0.29218, 0.30176, 0.31134, 0.32553,
                    0.34052, 0.37150, 0.41048, 0.42947, 0.43946, 0.44844, 0.46543, 0.48642, 0.51640, 0.55939,
                    0.62438], dtype='float32'),
                    axes=[np.arange(350, 701, 5.)]
                )
        self.Chi_tab = CLUT(np.array([
                    0.15300, 0.14900, 0.14400, 0.14000, 0.13600, 0.13100, 0.12700, 0.12300, 0.11900, 0.11800,
                    0.11748, 0.12066, 0.12259, 0.12326, 0.12269, 0.12086, 0.11779, 0.11372, 0.10963, 0.10560,
                    0.10165, 0.09776, 0.09393, 0.09018, 0.08649, 0.08287, 0.07932, 0.07584, 0.07242, 0.06907,
                    0.06579, 0.06257, 0.05943, 0.05635, 0.05341, 0.05072, 0.04829, 0.04611, 0.04419, 0.04253,
                    0.04111, 0.03996, 0.03900, 0.03750, 0.03600, 0.03400, 0.03300, 0.03280, 0.03250, 0.03300,
                    0.03400, 0.03500, 0.03600, 0.03750, 0.03850, 0.04000, 0.04200, 0.04300, 0.04400, 0.04450,
                    0.04500, 0.04600, 0.04750, 0.04900, 0.05150, 0.05200, 0.05050, 0.04400, 0.03900, 0.03400,
                    0.03000 ], dtype='float32'),
                    axes=[np.arange(350, 701, 5.)]
                )
        self.e_tab = CLUT(np.array([
                    0.77800, 0.76700, 0.75600, 0.73700, 0.72000, 0.70000, 0.68500, 0.67300, 0.67000, 0.66000,
                    0.64358, 0.64776, 0.65175, 0.65555, 0.65917, 0.66259, 0.66583, 0.66889, 0.67175, 0.67443,
                    0.67692, 0.67923, 0.68134, 0.68327, 0.68501, 0.68657, 0.68794, 0.68903, 0.68955, 0.68947,
                    0.68880, 0.68753, 0.68567, 0.68320, 0.68015, 0.67649, 0.67224, 0.66739, 0.66195, 0.65591,
                    0.64927, 0.64204, 0.64000, 0.63000, 0.62300, 0.61500, 0.61000, 0.61400, 0.61800, 0.62200,
                    0.62600, 0.63000, 0.63400, 0.63800, 0.64200, 0.64700, 0.65300, 0.65800, 0.66300, 0.66700,
                    0.67200, 0.67700, 0.68200, 0.68700, 0.69500, 0.69700, 0.69300, 0.66500, 0.64000, 0.62000,
                    0.60000], dtype='float32'),
                    axes=[np.arange(350, 701, 5.)]
                )
        self.bw_tab = CLUT(np.array([
                    0.0121, 0.0113, 0.0107, 0.0099, 0.0095, 0.0089, 0.0085, 0.0081, 0.0077, 0.0072,
                    0.0069, 0.0065, 0.0062, 0.0059, 0.0056, 0.0054, 0.0051, 0.0049, 0.0047, 0.0044,
                    0.0043, 0.0040, 0.0039, 0.0037, 0.0035, 0.0034, 0.0033, 0.0031, 0.0030, 0.0029,
                    0.0027] + map(lambda x: 0.0027*((x/500.)**-4.), np.arange(505, 701, 5.)),
                    dtype='float32'),
                    axes=[np.arange(350, 701, 5.)]
                )
        self.simspec = CLUT(np.array([
                4.953, 4.858, 4.734, 4.586, 4.432, 4.293, 4.177, 4.082, 4.017, 3.976, 3.949,
                3.939, 3.937, 3.974, 4.016, 4.046, 4.061, 4.015, 3.948, 3.862, 3.757, 3.621,
                3.466, 3.297, 3.118, 2.931, 2.754, 2.560, 2.350, 2.144, 1.937, 1.736, 1.551,
                1.393, 1.273, 1.185, 1.123, 1.080, 1.053, 1.032, 1.013, 1.001, 0.994, 1.012,
                1.029, 1.033, 1.016, 0.985, 0.971, 0.968, 0.972, 0.985, 1.000, 1.015, 1.029,
                1.046, 1.067, 1.087, 1.108, 1.127, 1.145, 1.159, 1.169, 1.173, 1.175, 1.171,
                1.159, 1.138, 1.098, 1.043, 0.980, 0.912, 0.846, 0.788, 0.742, 0.707, 0.678,
                0.658, 0.640, 0.627, 0.616, 0.603, 0.592, 0.579, 0.564, 0.553, 0.544, 0.534,
                0.523, 0.512, 0.501, 0.488, 0.476, 0.465, 0.454, 0.440, 0.431, 0.425, 0.419,
                0.413, 0.409], dtype='float32'),
                axes=[np.arange(650, 901, 2.5)])

        self.initialized = 0
        self.lam_join = 690.


    cdef init(self, float[:] wav, float sza, float vza, float raa):
        self.Nwav = len(wav)
        cdef int i
        cdef float lam

        if not self.initialized:
            self.wav   = np.zeros(self.Nwav+1, dtype='float32')
            self.Kw_i  = np.zeros(self.Nwav+1, dtype='float32')
            self.Chi_i = np.zeros(self.Nwav+1, dtype='float32')
            self.e_i   = np.zeros(self.Nwav+1, dtype='float32')
            self.bw_i  = np.zeros(self.Nwav+1, dtype='float32')
            self.simspec_i = np.zeros(self.Nwav+1, dtype='float32')
            self.Rw = np.zeros(self.Nwav, dtype='float32')
            self.initialized = 1

            if self.debug:
                self.bw  = np.zeros(self.Nwav, dtype='float32')
                self.atot  = np.zeros(self.Nwav, dtype='float32')
                self.bbtot  = np.zeros(self.Nwav, dtype='float32')
        elif len(wav) != len(self.Rw):
            raise Exception('Invalid length of wav')

        for i in range(self.Nwav):
            self.wav[i] = wav[i]
        self.wav[self.Nwav] = self.lam_join

        for i in range(self.Nwav+1):
            # FIXME: this loop can be made faster
            lam = self.wav[i]

            self.Kw_tab.lookup(0, lam)
            self.Kw_i[i] = self.Kw_tab.interp()

            self.Chi_tab.lookup(0, lam)
            self.Chi_i[i] = self.Chi_tab.interp()

            self.e_tab.lookup(0, lam)
            self.e_i[i] = self.e_tab.interp()

            self.bw_tab.lookup(0, lam)
            self.bw_i[i] = self.bw_tab.interp()

            if self.simspec.lookup(0, lam) == 0:
                self.simspec_i[i] = self.simspec.interp()


    cdef float[:] calc_rho(self, float[:] x):
        cdef float rw_join

        # wavelength loop: visible
        for i in range(self.Nwav):
            self.Rw[i] = self.calc_rho_vis(i, x)

        # towards NIR
        rw_join = self.calc_rho_vis(self.Nwav, x)

        # wavelength loop: NIR
        for i in range(self.Nwav):
            if self.Rw[i] < 0:
                self.Rw[i] = self.calc_rho_nir(i, rw_join)

        return self.Rw

    cdef float calc_rho_vis(self, int i, float[:] x):
        '''
        reflectance calculation for visible bands
        return -1 if invalid wavelength
        '''

        cdef float Kw
        cdef float Chi, e, lam
        cdef float Kbio
        cdef float logchl = x[0]
        cdef float bbs0 = x[1]
        cdef float bbs, bp550, bbp, bb, bw
        cdef float ays0, a_ys
        cdef float Sys, Snap
        cdef float v
        cdef int j
        cdef float bbs_spec
        ays0 = 0.

        Kw  = self.Kw_i[i]
        Chi = self.Chi_i[i]
        e   = self.e_i[i]
        bw  = self.bw_i[i]
        lam = self.wav[i]

        if lam > 700:
            return -1

        Kbio = Chi * (10.**(e * logchl))

        if (logchl < 0.301029995664):
            v = 0.5 * (logchl - 0.3) # // 0.301029995664 == log10(2)
        else:
            v = 0;

        bbs_spec = -1.
        bbs = bbs0*((lam/550.)**bbs_spec)

        bp550 = 0.416 * (10.**(0.766 * logchl))
        bbp = (0.002 + 0.01*(0.5 - 0.25*logchl)*((lam/550.)**v)) * bp550
        bb = 0.5 * bw + bbp + bbs

        # absorption by yellow substances
        Sys = 0.014
        a_ys = ays0*exp(-Sys*(lam - 410.))

        Kd = Kw + Kbio + bbs + a_ys

        # absorption by detritus (non-algal particles)
        Snap = 0.011

        # Correction of detrital absorption for clear waters
        # (see Morel et al, 2007, "Optical properties of the 'clearest' natural waters")
        a_nap = - exp(-7.55*(10**(logchl*0.08)))*exp(-Snap*(lam - 420.))


        # iterations
        u = 0.75
        for j in range(3):
            a = u * Kd
            R = 0.33 * bb / (a + a_ys + a_nap)

            u = 0.90 * 1/(1 + 2.25*R) * (1 - R)

        # TODO:
        # interpolate f/Q coefficients

        if self.debug:
            self.bw[i] = bw
            self.bbtot[i] = bb
            self.atot[i] = a + a_ys + a_nap

        return R * 0.544

    cdef float calc_rho_nir(self, int i, float rw_join):
        return self.simspec_i[i]*rw_join/self.simspec_i[self.Nwav]

    def calc(self, w, logchl, bbs=0., sza=0., vza=0., raa=0.):
        '''
        water reflectance calculation (python interface)
        w: wavelength (nm) [float, list or array]
        returns above-water reflectance as float, list or array
        '''
        if isinstance(w, np.ndarray):
            self.out_type = np.array
            wav = w.astype('float32')
        elif isinstance(w, list):
            self.out_type = list
            wav = np.array(w, dtype='float32')
        else: # float
            self.out_type = lambda x: x[0]
            wav = np.array([w], dtype='float32')

        self.init(wav, float(sza), float(vza), float(raa))
        params = np.zeros(2, dtype='float32')
        params[0] = logchl
        params[1] = bbs
        R = self.calc_rho(params)
        return self.out_type(R)

    def iops(self):
        '''
        returns the IOPs for the previously calculated spectrum
        '''
        assert self.debug

        return {
                'bw':    self.out_type(self.bw),
                'atot':  self.out_type(self.atot),
                'bbtot': self.out_type(self.bbtot),
                }


def test():
    pr = ParkRuddick('auxdata/common/')
    pr.init(np.linspace(401, 800, 100, dtype='float32'), 0, 0, 0)
    a = pr.calc_rho(np.array([0., 0.], dtype='float32'))
    # print np.array(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from os import environ
from pathlib import Path
from dotenv import load_dotenv
import warnings

'''
SMART-G constant variables

'''

DIR_ROOT = Path(__file__).resolve().parent.parent
NPSTK = 4
dir_auxdata_old = DIR_ROOT / 'auxdata'

load_dotenv(DIR_ROOT / '.env') # To consider .env file

# Allow old way in v1.1, but remove it in next releases
try:
    dir_auxdata_new = environ['SMARTG_DIR_AUXDATA']
    DIR_AUXDATA = Path(dir_auxdata_new)
except KeyError:
    if not dir_auxdata_old.is_dir():
        raise NameError("The environment variable 'SMARTG_DIR_AUXDATA' does not exist!")
    else:
        warnings.simplefilter('always', DeprecationWarning)
        warn_message = "\nThe 'SMARTG_DIR_AUXDATA' variable does not exist, but the folder auxdata has \n" + \
                       "been found in the root directory and is used instead. This is deprecated, and will \n" + \
                       "not be allowed in next releases. Please give the auxdata directory path by \n" + \
                       "using the environment variable 'SMARTG_DIR_AUXDATA' instead."
        warnings.warn(warn_message, DeprecationWarning)
        DIR_AUXDATA = dir_auxdata_old

# TODO for constant variable must use uppercase 
dir_libradtran = DIR_AUXDATA / 'libRadtran'
dir_libradtran_atmmod = dir_libradtran / 'data' / 'atmmod'


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This module is deprecated. It will be removed in one of the next versions of smartg.
# All the classes/functions are now available in the geoclide package.

import math 
import numpy as np
from warnings import warn

#####################################################################################
class Vector(object):
    '''
    Definition of Vector

    Creation of the class vector

    args : x, y and z directions
    '''
    def __init__(self, x = 0, y = 0, z = 0):
        warn_message = 'Using Vector from smartg.geometry is depracated as of smartg v1.1.0. ' + \
            'Use geoclide package instead.\n For example: import geoclide as gc; gc.Vector().'
        warn(warn_message, DeprecationWarning)
        if isinstance(x, Vector) or isinstance(x, Point) or isinstance(x, Normal):
            self.x = x.x; self.y = x.y; self.z = x.z;
        elif isinstance(x, np.ndarray):
            self.x = x[0]; self.y = x[1]; self.z = x[2]
        else: 
            self.x = x; self.y = y; self.z = z;

    def __eq__(self, v2):
        if isinstance(v2, Vector):
            return (self.x==v2.x) and (self.y==v2.y) and (self.z==v2.z)
        else:
            raise NameError('Equality with a Vector must be with another Vector')

    def __add__(self, v2):
        if isinstance(v2, Vector):
            return Vector(self.x+v2.x, self.y+v2.y, self.z+v2.z) 
        else:
            raise NameError('Addition with a Vector must be with another Vector')

    def __sub__(self, v2):
        if isinstance(v2, Vector):
            return Vector(self.x-v2.x, self.y-v2.y, self.z-v2.z)
        else:
            raise NameError('Substraction with a Vector must be with another Vector')

    def __truediv__(self, sca):
        if (type(sca) != Vector) and (type(sca) != Point) and (type(sca) != Normal):
            return Vector(self.x/sca, self.y/sca, self.z/sca) 
        else:
            raise NameError('div accepted only with scalar')
    def __mul__(self, sca): 
        if (type(sca) != Vector) and (type(sca) != Point) and (type(sca) != Normal):
            return Vector(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise NameError('mul accepted only with scalar')

    def __getitem__(self, ind):
        if ind == 0:
            return self.x;
        elif ind == 1:
            return self.y;
        elif ind == 2 :
            return self.z;
        else:
            raise NameError('Indice out of range!')

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'

    def Length(self):
        return math.sqrt(self.x*self.x + self.y*self.y + self.z*self.z) # L2 norm

    def asarr(self):
        return np.array([self.x, self.y, self.z])
    
#####################################################################################

#####################################################################################
class Point(object):
    '''
    Definition of Point

    Creation of the class Point

    args : x, y and z coordinates
    '''
    def __init__(self, x = 0, y = 0, z = 0):
        warn_message = 'Using Point from smartg.geometry is depracated as of smartg v1.1.0. ' + \
            'Use geoclide package instead.\n For example: import geoclide as gc; gc.Point().'
        warn(warn_message, DeprecationWarning)
        if isinstance(x, Vector) or isinstance(x, Point) or isinstance(x, Normal):
            self.x = x.x; self.y = x.y; self.z = x.z;
        else:
            self.x = x; self.y = y; self.z = z;

    def __eq__(self, p2):
        if isinstance(p2, Point):
            return (self.x==p2.x) and (self.y==p2.y) and (self.z==p2.z)
        else:
            raise NameError('Equality with a Point must be with another Point')

    def __add__(self, v):
        if isinstance(v, Vector):
            return Point(self.x+v.x, self.y+v.y, self.z+v.z)
        else:
            raise NameError('Addition with a Point must be with a Vector')

    def __sub__(self, vp2):
        if isinstance(vp2, Vector):
            return Point(self.x-vp2.x, self.y-vp2.y, self.z-vp2.z)
        elif isinstance(vp2, Point):
            return Vector(self.x-vp2.x, self.y-vp2.y, self.z-vp2.z)
        else:
            raise NameError('Subs with a Point must be with another Point or a Vector')

    def __truediv__(self, sca):
        if (type(sca) != Vector) and (type(sca) != Point) and (type(sca) != Normal):
            return Point(self.x/sca, self.y/sca, self.z/sca)
        else:
            raise NameError('div accepted only with scalar')

    def __mul__(self, sca):
        if (type(sca) != Vector) and (type(sca) != Point) and (type(sca) != Normal):
            return Point(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise NameError('mul accepted only with scalar')

    def __getitem__(self, ind):
        if ind == 0:
            return self.x;
        elif ind == 1:
            return self.y;
        elif ind == 2 :
            return self.z;
        else:
            raise NameError('indice out of range!')

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'
#####################################################################################

#####################################################################################
class Normal(object):
    '''
    Definition of Normal

    Creation of the class Normal

    args : x, y and z directions
    '''
    def __init__(self, x = 0, y = 0, z = 0):
        warn_message = 'Using Normal from smartg.geometry is depracated as of smartg v1.1.0. ' + \
            'Use geoclide package instead.\n For example: import geoclide as gc; gc.Normal().'
        warn(warn_message, DeprecationWarning)
        if isinstance(x, Vector) or isinstance(x, Point) or isinstance(x, Normal):
            self.x = x.x; self.y = x.y; self.z = x.z;
        else:
            self.x = x; self.y = y; self.z = z;

    def __eq__(self, n2):
        if isinstance(n2, Normal):
            return (self.x==n2.x) and (self.y==n2.y) and (self.z==n2.z)
        else:
            raise NameError('Equality with a Normal must be with another Normal')

    def __add__(self, n2):
        if isinstance(n2, Normal):
            return Normal(self.x+n2.x, self.y+n2.y, self.z+n2.z) 
        else:
            raise NameError('Addition with a Normal must be with another Normal')

    def __sub__(self, n2):
        if isinstance(n2, Normal):
            return Vector(self.x-n2.x, self.y-n2.y, self.z-n2.z)
        elif n2 == 0:
            return (self.x==-1*self.x) and (self.y==-1*self.y) and (self.z==-1*self.z)
        else:
            raise NameError('Substraction with a Normal must be with another Normal')

    def __truediv__(self, sca):
        if (type(sca) != Vector) and (type(sca) != Point) and (type(sca) != Normal):
            return Vector(self.x/sca, self.y/sca, self.z/sca) 
        else:
            raise NameError('div accepted only with scalar')

    def __mul__(self, sca): 
        if (type(sca) != Vector) and (type(sca) != Point) and (type(sca) != Normal):
            return Vector(sca*self.x, sca*self.y, sca*self.z)
        else:
            raise NameError('mul accepted only with scalar')

    def __getitem__(self, ind):
        if ind == 0:
            return self.x;
        elif ind == 1:
            return self.y;
        elif ind == 2 :
            return self.z;
        else:
            raise NameError('Indice out of range!')

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.z) + ')'

    def Length(self):
        return math.sqrt(self.x*self.x + self.y*self.y + self.z*self.z) # L2 norm
#####################################################################################

#####################################################################################
class Ray(object):
    '''
    Definition of Ray

    Creation of the class Ray

    o     : Reference point of the ray (Point class)
    d     : Direction of the ray (Vector class)
    t     : Initial time for the initial position of the ray
            t = 0 means o = the initial position
    start : Min value of t
    end   : Max value of t

    '''
    def __init__(self, o = Point(), d = Vector(), t = 0, start = 0, end = float("inf")):
        warn_message = 'Using Ray from smartg.geometry is depracated as of smartg v1.1.0. ' + \
            'Use geoclide package instead.\n For example: import geoclide as gc; gc.Ray().'
        warn(warn_message, DeprecationWarning)
        if isinstance(o, Ray):
            self.o = o.o
            self.d = o.d
            self.t = o.t
            self.mint = o.mint
            self.maxt = o.maxt
        else:
            self.o = o
            self.d = d
            self.t = t
            self.mint = start
            self.maxt = end

    def __getitem__(self, t):
        if (type(t) != Vector) and (type(t) != Point) and (type(t) != Normal):
            return (Point(self.o) + Vector(self.d * t))
        else:
            raise NameError('t must be a scalar')

#####################################################################################

#####################################################################################
class BBox(object):
    '''
    Definition of Bounding Box

    Creation of the bounding box class thanks to 2 points
    
    p1 : First point (Point class)
    p2 : Second point (Point class)
    '''
    def __init__(self, p1 = None, p2 = None):
        warn_message = 'Using BBox from smartg.geometry is depracated as of smartg v1.1.0. ' + \
            'Use geoclide package instead.\n For example: import geoclide as gc; gc.BBox().'
        warn(warn_message, DeprecationWarning)
        if (isinstance(p1, Point) or (p1 is None)) and \
           (isinstance(p2, Point) or (p2 is None)):
            if ((p1 is None) and (p2 is None)):
                self.pMax = Point(float("inf"), float("inf"), float("inf"))
                self.pMin = Point(float("-inf"), float("-inf"), float("-inf"))
            elif (p2 is None):
                self.pMax = Point(p1)
                self.pMin = Point(p1)    
            elif (p1 is None):
                self.pMax = Point(p2)
                self.pMin = Point(p2) 
            else:
                self.pMax = Point(max(p1.x, p2.x), max(p1.y, p2.y), max(p1.z, p2.z))
                self.pMin = Point(min(p1.x, p2.x), min(p1.y, p2.y), min(p1.z, p2.z))
        else:
            raise NameError('Bounding Box constructor accepts only points')
        
    def union(self, arg1, arg2):
        if isinstance(arg1, BBox):
            ret = BBox()
            if isinstance(arg2, Point):
                ret.pMin.x = min(arg1.pMin.x, arg2.x);
                ret.pMin.y = min(arg1.pMin.y, arg2.y);
                ret.pMin.z = min(arg1.pMin.z, arg2.z);
                ret.pMax.x = max(arg1.pMax.x, arg2.x);
                ret.pMax.y = max(arg1.pMax.y, arg2.y);
                ret.pMax.z = max(arg1.pMax.z, arg2.z);
            elif isinstance(arg2, BBox):
                ret.pMin.x = min(b.pMin.x, arg2.pMin.x);
                ret.pMin.y = min(b.pMin.y, arg2.pMin.y);
                ret.pMin.z = min(b.pMin.z, arg2.pMin.z);
                ret.pMax.x = max(b.pMax.x, arg2.pMax.x);
                ret.pMax.y = max(b.pMax.y, arg2.pMax.y);
                ret.pMax.z = max(b.pMax.z, arg2.pMax.z);
            else:
                raise NameError('arg2 of BBox union have to be BBox or Point class')
            return ret
        else:
            raise NameError('arg1 of BBox union have to be BBox class')            
            
    def get_vertices(self):
        '''
        get the 8 vertices of a BBox as a list of points.
        p0==pMin, then next 3 points are in the plane
        pMin.z,  the order being anti-clockwise.
        Next 4 points are in the plane pMax.z, starting with point
        just above p0, so pMax==p6
        '''
        return [
            Point(self.pMin.x,self.pMin.y,self.pMin.z),
            Point(self.pMax.x,self.pMin.y,self.pMin.z),
            Point(self.pMax.x,self.pMax.y,self.pMin.z),
            Point(self.pMin.x,self.pMax.y,self.pMin.z),
            Point(self.pMin.x,self.pMin.y,self.pMax.z),
            Point(self.pMax.x,self.pMin.y,self.pMax.z),
            Point(self.pMax.x,self.pMax.y,self.pMax.z),
            Point(self.pMin.x,self.pMax.y,self.pMax.z)
        ]

    def include_point(self, P):
        '''
        test if Point P is included in BBox
        '''
        return (P.x >= self.pMin.x) and (P.x <= self.pMax.x) and \
               (P.y >= self.pMin.y) and (P.y <= self.pMax.y) and \
               (P.z >= self.pMin.z) and (P.z <= self.pMax.z)


    def IntersectP(self, Ray) :
        '''
        test if Ray Ray intersects the BBox and finds both intersection points
        '''
        t0 = 0.
        epsi = 1e-32 * 0.5
        t1 = np.inf
        gamma3 = (3*epsi)/(1 - 3*epsi)
        for i in range(3):
            if Ray.d[i]!= 0 :
                invRayDir = 1. / Ray.d[i]
            else: invRayDir=1e32
            tNear = (self.pMin[i] - Ray.o[i]) * invRayDir
            tFar  = (self.pMax[i] - Ray.o[i]) * invRayDir

            if (tNear > tFar):
                tmp  = tNear
                tNear= tFar
                tFar = tmp
            tFar *= 1 + 2*gamma3
            t0 = tNear if tNear > t0 else t0
            t1 = tFar  if  tFar < t1 else t1
            if (t0 > t1) : return 0, 0, False
        return t0, t1, True


def CommonVertices(BBox1, BBox2):
    ''' 
    return a list of boolean checking if vertices are common between BBoxes 
    '''
    warn_message = 'Using CommonVertices from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.get_common_vertices().'
    warn(warn_message, DeprecationWarning)
    return np.array(list((map(lambda x: x in BBox2.get_vertices(), BBox1.get_vertices()))))

def CommonFace(BBox1, BBox2, Fill_value=None):
    ''' 
    return the face index of th BBox1 which is common to BBox2 with
    the convention of index from 0 to 5, for +X,-X,+Y,-Y,+Z,-Z faces
    # (en.wikipedia.org/wiki/Cube_mapping)
    return Fill if there is no common face
    '''
    warn_message = 'Using CommonFace from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.get_common_face().'
    warn(warn_message, DeprecationWarning)
    ok = CommonVertices(BBox1, BBox2)
    if ok.sum()==4:
        n  = np.arange(8)[ok]
        if   np.array_equal(n, np.array([1,2,5,6])):
            return 0
        elif np.array_equal(n, np.array([0,3,4,7])):
            return 1
        elif np.array_equal(n, np.array([2,3,6,7])):
            return 2
        elif np.array_equal(n, np.array([0,1,4,5])):
            return 3
        elif np.array_equal(n, np.array([4,5,6,7])):
            return 4
        elif np.array_equal(n, np.array([0,1,2,3])):
            return 5
        else: return Fill_value

    else :
        return Fill_value

#####################################################################################
def Dot(a, b):
    warn_message = 'Using Dot from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.dot().'
    warn(warn_message, DeprecationWarning)
    if (isinstance(a, Vector) or isinstance(a, Normal)) and \
       (isinstance(b, Vector) or isinstance(b, Normal)):
        return (a.x*b.x + a.y*b.y + a.z*b.z)
    else:
        raise NameError('Dot argments have to be Vector or Normal classes')

def Cross(a, b):
    warn_message = 'Using Cross from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.cross().'
    warn(warn_message, DeprecationWarning)
    if (isinstance(a, Vector) and isinstance(b, Vector)) or \
    (isinstance(a, Vector) and isinstance(b, Normal)) or \
    (isinstance(a, Normal) and isinstance(b, Vector)):
        return Vector((a.y*b.z)-(a.z*b.y), (a.z*b.x)-(a.x*b.z), (a.x*b.y)-(a.y*b.x))
    elif isinstance(a, Normal) and isinstance(b, Normal):
        raise NameError('Only 1 Normal (Cross args) is tolerate not 2')
    else:
        raise NameError('Cross args must be Vector or Normal')

def Normalize(v):
    warn_message = 'Using Normalize from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.normalize().'
    warn(warn_message, DeprecationWarning)
    if isinstance(v, Vector) or isinstance(v, Normal):
        return v / v.Length()
    else:
        raise NameError('Normalize argument have to be Vector or Normal class')

def CoordinateSystem(v1):
    '''
    Create an orthogonal coordinate system from 1 vector (v1)
    '''
    warn_message = 'Using CoordinateSystem from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.coordinate_system().'
    warn(warn_message, DeprecationWarning)
    if (abs(v1.x) > abs(v1.y)):
        invLen = 1/ math.sqrt(v1.x*v1.x + v1.z*v1.z)
        v2 = Vector(-v1.z*invLen, 0, v1.x*invLen)
    else:
        invLen = 1/ math.sqrt(v1.y*v1.y + v1.z*v1.z)
        v2 = Vector(0, v1.z*invLen, -v1.y*invLen)
    v3 = Cross(v1, v2)
    return v2, v3

def Distance(p1, p2):
    warn_message = 'Using Distance from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.distance().'
    warn(warn_message, DeprecationWarning)
    if isinstance(p1, Point) and isinstance(p2, Point):
        return (p1 - p2).Length()
    else:
        raise NameError('Distance arguments have to be Point classes)')

def FaceForward(a, b):
    '''
    Flip the Vector/Normal a if the Vector/Normal b is in the opposite direction.
    For exemple, it can be useful to flip a surface normal so that it lies in the
    same hemisphere as a given vector.
    Args : Vector or Normal a, b
    Output : Possibly fliped Vector or Normal a
    '''
    warn_message = 'Using FaceForward from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.face_forward().'
    warn(warn_message, DeprecationWarning)
    if (isinstance(a, Vector) or isinstance(a, Normal)) and \
    (isinstance(b, Vector) or isinstance(b, Normal)):
        return (a*-1) if (Dot(a, b) < 0) else a
    else:
        raise NameError('FaceForward args have to be Vector or Normal classes')

def rotation3D(theta, u):
    ''' 
    rotation matrix of an angle theta in degree around unit vector u
    '''
    warn_message = 'Using rotation3D from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.get_rotate_tf(). ' + \
                   'Note that gc.get_rotate_tf() returns 4x4 matrix instead of 3x3 matrix with rotation3D.'
    warn(warn_message, DeprecationWarning)
    # Rodrigues rotation formula
    ct = np.cos(np.radians(theta))
    st = np.sqrt(1-ct*ct)
    A = np.zeros((3,3))
    for k in range(3) : A[k,k] = 1.
    B = np.zeros((3,3))
    B[0,1] = -u.z
    B[0,2] =  u.y
    B[1,0] =  u.z
    B[1,2] = -u.x
    B[2,0] = -u.y
    B[2,1] =  u.x
    C = B.dot(B)
    return A + B*st + C*(1-ct)


def MaxDimension(v):
    warn_message = 'Using MaxDimension from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.vmax().'
    warn(warn_message, DeprecationWarning)
    if isinstance(v, Vector):
        val = 0
        if (v.x > v.y):
            if(v.x > v.z):
                val = 0
            else:
                val = 2
        else:
            if(v.y > v.z):
                val = 1
            else:
                val = 2
        return int(val)
    else:
        raise NameError('v argument must be a Vector')
        
def Permute(v, xx, yy, zz):
    warn_message = 'Using Permute from smartg.geometry is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n For example: import geoclide as gc; gc.permute().'
    warn(warn_message, DeprecationWarning)
    if isinstance(v, Vector):
        return Vector(v[xx], v[yy], v[zz])
    else:
        raise NameError('v argument must be a Vector')

if __name__ == '__main__':
    
    a = Vector(10, 2, 3)
    b = Point(2, 4, 6)
    c = Vector(b)*2
    p1 = Point(4, 8, 12)
    print("a =", a)
    print("b =", b)
    print("c =", c)
    print("Dot(a, c) =", Dot(a, c))
    print("Dot(a, b) =", Dot(a, Vector(b)))
    print("Cross(a, c) =", Cross(a, c))
    print("Length(a) =", a.Length())
    print("normalize(a) =", Normalize(a))
    print("Length of Normalize(a) =", Normalize(a).Length())

    v1 = Vector(2, 1, 0)
    v2, v3 = CoordinateSystem(v1)

    print("v1 =", v1)
    print("v2 =", v2)
    print("v3 =", v3)
    print("p2 =", p1-b)
    print("type(p2) =", type(p1-b))
    print("Distance(b, p1) =",  Distance(b, p1))
    print("a*2 =", a*2)

    vec1 = Vector (-2, 4, -1)
    vec2 = Vector (5, 1, 3)

    print("FaceForward(vec1, vec2) =", FaceForward(vec1, vec2))

    rayon = Ray(p1, vec2)
    print("p1 =", rayon.o)
    print("vec2 =", rayon.d)
    print("rayon(t=1) =", rayon[1])
    print("rayon(t=2) =", rayon[2])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import
import numpy as np
from luts.luts import LUT, Idx

class Albedo_cst(object):
    '''
    Constant albedo (white)
    '''
    def __init__(self, alb):
        self.alb = alb

    def get(self, wl):
        alb = np.zeros(np.array(wl).shape, dtype='float32')
        alb[...] = self.alb
        return alb

class Albedo_speclib(object):
    '''
    Albedo from speclib
    (http://speclib.jpl.nasa.gov/)
    '''
    def __init__(self, filename):
        data = np.genfromtxt(filename, skip_header=26)
        # convert X axis from micrometers to nm
        # convert Y axis from percent to dimensionless
        self.data = LUT(data[:,1]/100., axes=[data[:,0]*1000.], names=['wavelength'])

    def get(self, wl):
        return self.data[Idx(wl, fill_value='extrapolate')]

class Albedo_spectrum(object):
    '''
    Albedo R(lambda)

    R spectral albedo, lam in nm
    '''
    def __init__(self, R, lam):
        self.data = LUT(R, axes=[lam], names=['wavelength'])

    def get(self, wl):
        return self.data[Idx(wl, fill_value='extrapolate')]


class Albedo_map(object):
    '''
    Albedo map of Albedo objects:
    A 2D horizontal map of spectral albedos can be constructed
    Spectral albedos are limited to a MAX_NREF=10 different kind, could be extended
    They should be defined using Albedo_cst, Albedo_spectrum or Albedo_speclib classes

    The horizontal grid for the 2D map of albedo is rectangular
    and x and y boundaries on the surface (in km) are encoded in monotonic np.arrays whose values 
    are the upper limit of the rectangles:
    if x = [x0, x1, x2, ..., xn], the the limits are [-Inf, x0], [x0, x1], ..., [xn-1, xn], 
    with xn big enough to be considered as +Inf 

    Albedo index (2D), x in km, y in km, Alist: list of Albedo objects
    we assigned each rectangle an index in Alb_list, in a 2D array of shape (len(x),len(y))
    Negative indices are for the surf properties.
    '''
    def __init__(self, Ai, x, y, Alist):
        self.map = LUT(Ai, axes=[x, y], names=['X', 'Y'])
        self.list = Alist
        self.NALB = len(Alist)

    def get(self, wl):
        return np.stack([ALB.get(wl) for ALB in self.list]).T

    def get_map(self, x0, y0):
        return self.map[Idx(x0, round=True, fill_value='extrema'),
                        Idx(y0, round=True, fill_value='extrema')].astype(int)





import numpy as np

def Gauss(ks, Aj, kj, Dkj):

    return Aj * 1./Dkj * np.exp(-4*np.log(2)*(ks-kj)**2/Dkj**2)

def fR(ks):
    A  = np.array([0.41, 0.39, 0.10, 0.10])
    k  = np.array([3250., 3425., 3530., 3625.])
    Dk = np.array([210., 175., 140., 140.])
    norm = np.sum(A) * np.sqrt(np.pi/4/np.log(2))
    norm = 1./norm
    Su=np.zeros_like(ks)
    for j in range(4):
        Su+= Gauss(ks, A[j], k[j], Dk[j])

    return Su*norm


def V2d(lam, Nl=16):
    '''
    Ocean Vibrational Raman Spectrum

    lam central wavelength in nm
    '''
    k   = 1e7/lam   # cm-1
    k0  = k - 2950. # cm-1
    k1  = k - 3850. # cm-1
    w0  = 1e7/k0
    w1  = 1e7/k1
    wgrid = np.linspace(w0, w1, num=Nl, dtype=np.float32)
    ks    = 1e7*(1./lam[np.newaxis,:]-1./wgrid)
    
    return wgrid, 1e7/wgrid**2 * fR(ks)


def V2d_inv(lam, Nl=16):
    '''
    Inverse Ocean Vibrational Raman Spectrum

    lam central wavelength in nm
    '''
    k   = 1e7/lam # cm-1
    k0  = k + 3850. # cm-1
    k1  = k + 2950. # cm-1
    w0  = 1e7/k0
    w1  = 1e7/k1
    wgrid = np.linspace(w0, w1, num=Nl, dtype=np.float32).T
    ks    = 1e7*(1./wgrid - 1./lam[:,np.newaxis])
    
    return wgrid, 1e7/wgrid**2 * fR(ks)



#!/usr/bin/env python
# vim:fileencoding=utf-8



from numpy import loadtxt, zeros, NaN, interp
from pathlib import Path


def a_w(lam):
    '''
    pure water absorption from Pope and Fry + Palmer and Williams
    '''

    this_dir = Path(__file__).parent
    file_pope97 = this_dir / 'data' / 'pope97.dat'
    file_palmer = this_dir / 'data' / 'palmer74.dat'

    # 1) Pope&Fry
    wl_popefry = loadtxt(file_pope97, skiprows=6, usecols=(0,))
    aw_popefry = loadtxt(file_pope97, skiprows=6, usecols=(1,))
    aw_popefry *= 100. # convert cm-1 => m-1

    # 2) Palmer&Williams
    wl_palmerw = loadtxt(file_palmer, skiprows=5, usecols=(0,))[::-1]
    aw_palmerw = loadtxt(file_palmer, skiprows=5, usecols=(1,))[::-1]
    aw_palmerw *= 100. # convert cm-1 => m-1

    if not isinstance(lam,float) :
        aw = zeros(lam.shape, dtype='float32') + NaN
        pf = lam < 715.  # use pope&fry
        aw[pf] = interp(lam[pf], wl_popefry, aw_popefry, right=NaN, left=NaN)
        aw[~pf] = interp(lam[~pf], wl_palmerw, aw_palmerw, right=NaN, left=NaN)
        return aw
    else :
        if lam<715 : return interp(lam, wl_popefry, aw_popefry, right=NaN, left=NaN)
        else: return interp(lam, wl_palmerw, aw_palmerw, right=NaN, left=NaN)





#!/usr/bin/env python
# vim:fileencoding=utf-8

'''
Calculate the ocean water absorption, scattering coefficients and phase
function using the wavelength and the SPM concentration
'''

from numpy import sin, cos, pi, exp
from numpy import arange, zeros, log10
from seawater_scattering_Zhang2009 import swscat
from water_absorption import a_w
from phase_functions import fournierForand, PhaseFunction, fournierForandB
from iop import IOP



class IOP_SPM(IOP):
    '''
    Initialize the IOP model (SPM dominated waters)

    Arguments:
        SPM: suspended particulate matter in g/m3
        NANG: number of angles for Phase function
        ang_trunc : truncature angle for Fournier Forand Phase function
        gamma: is the spectral dependency of the particulate backscattering
        alpha: parameter for CDOM absorption
        nbp: refractive index of particles (relative to water)
        ALB: albedo of the sea floor
        DEPTH: depth in m , default None = Semi infinite
        pfwav: list of arrays at which the phase functions are calculated
    '''
    def __init__(self, SPM, NANG=72001, ang_trunc=5., gamma=0.5,
            alpha=1., nbp=1.15, ALB=0., DEPTH=None, pfwav=None, verbose=False):
        self.__SPM = SPM
        self.__NANG = NANG
        self.__ang_trunc = ang_trunc
        self.__gamma = gamma
        self.__alpha = alpha
        self.__nbp = nbp
        self.alb = ALB
        self.depth = DEPTH
        self.pfwav=pfwav
        self.verbose = verbose
        self.last = None   # stores last result (managed by parent class)


    def calc(self, w, skip_phase=False):
        '''
        Calculate atot, btot and phase function (monochromatic)

        Arguments:
            w: wavelength in nm

        Returns (atot, btot, phase) where:
            * atot is the total absorption coefficient
            * btot is the total scattering coefficient
            * phase is the PhaseFunction object
        '''

        SPM = self.__SPM
        NANG = self.__NANG
        ang_trunc = self.__ang_trunc
        nbp = self.__nbp
        gamma = self.__gamma
        alpha = self.__alpha

        # pure sea water scattering 
        bw = swscat(w)

        # particulate backscattering
        bbp650 = 10**(1.03*log10(SPM) - 2.06) # Neukermans et al 2012
        bbp = bbp650*(w/650.)**(-gamma)

        # CDM absorption
        aCDM = alpha*0.031*SPM*exp(-0.0123*(w-443.))

        # pure sea water absorption
        aw = a_w(w)

        # Backscattering ratio of particles (non-troncated)
        Bp = fournierForandB(nbp, 3.+gamma)
        # Scattering coefficient of particles
        bp   = bbp/Bp


        #
        # total absorption
        atot = aCDM + aw

        if skip_phase:
            P0, P1 = None, None
        else:
            #
            # phase function
            #
            ang = pi * arange(NANG, dtype='float64')/(NANG-1)    # angle in radians

            # pure water
            pf0 = zeros((NANG, 4), dtype='float64') # pure water phase function
            pf0[:,1] = 0.75
            pf0[:,0] = 0.75 * cos(ang)**2
            pf0[:,2] = 0.75 * cos(ang)
            pf0[:,3] = 0.
            P0 = PhaseFunction(ang, 2*pf0, degrees=False)

            # particles (troncature)
            itronc = int(NANG * ang_trunc/180.)
            pf1 = zeros((NANG, 4), dtype='float64') # pure water phase function
            # assuming that the slope of Junge power law mu and slope of spectral dependence of scattering is mu=3+gamma
            pf1[itronc:,0] = 0.5*fournierForand(ang[itronc:],nbp,3.+gamma)
            pf1[:itronc,0] = 0.5*fournierForand(ang[itronc ],nbp,3.+gamma) 
            pf1[:,1] = pf1[:,0]
            pf1[:,2] = 0.
            pf1[:,3] = 0.

            # normalization after truncation
            integ_ff = 0.
            integ_ff_back = 0.
            for iang in xrange(1, NANG):
                dtheta = ang[iang] - ang[iang-1]
                pm1 = pf1[iang-1,0] + pf1[iang-1,1]
                pm2 = pf1[iang,0] + pf1[iang,1]
                sin1 = sin(ang[iang-1])
                sin2 = sin(ang[iang])
                integ_ff += dtheta*((sin1*pm1+sin2*pm2)/3. + (sin1*pm2+sin2*pm1)/6.)
                if ang[iang]>pi/2. :
                    integ_ff_back += dtheta*((sin1*pm1+sin2*pm2)/3. + (sin1*pm2+sin2*pm1)/6.)
            rat1 = integ_ff/2.
            pf1 *= 1/rat1
            P1 = PhaseFunction(ang, 2*pf1, degrees=False, coef_trunc=rat1)

        return atot, [(bw, P0), (bp, P1)]

    def __str__(self):
        return 'SPM={}'.format(self.__SPM)





#!/usr/bin/env python
# encoding: utf-8


import numpy as np
import tempfile


class IOP(object):
    '''
    abstract base class for IOP models
    '''
    def write(self, wl,  dir_profile, dir_phases, dir_list_phases, Zbottom=10000., Nlayer=2):
        '''
        write profiles and phase functions at bands wl (list)
        returns a tuple (profile, phases) where
        - profiles is the filename containing the concatenated profiles
        - phase is a file containing the list pf phase functions
        -Zbottom is the depth of the bottom in m
        -Nlayer is the number of layer in the ocean
        '''
        # convert to list if wl is a scalar
        if isinstance(wl, (float, int)):
            wl = [wl]

        iops, phases = self.calc_bands(wl)

        # write the profiles
        profil_oce = tempfile.mktemp(dir=dir_profile, prefix='profil_oce_')
        f = open(profil_oce, 'w')
        for (atot, btot, ipha) in iops:
                f.write('# I   DEPTH    H(I)    SSA(I)  IPHA\n')
                f.write('0 0. 0. 1. 0\n')
                for n in range(Nlayer):
                    z= (n+1.)*Zbottom/Nlayer
                    tau=z*(atot+btot) 
                    f.write('{} {} {} {} {}\n'.format(n+1,z,-tau,btot/(atot+btot), ipha))
                #f.write('1 1000. -1.e10 {} {}\n'.format(btot/(atot+btot), ipha))
        f.close()

        # write the phase functions and list of phase functions
        file_list_pf_ocean = tempfile.mktemp(dir=dir_list_phases, prefix='list_pf_ocean_')
        fp = open(file_list_pf_ocean, 'w')
        for p in phases:
            file_phase = tempfile.mktemp(dir=dir_phases, prefix='pf_ocean_')
            p.write(file_phase)
            fp.write(file_phase+'\n')
        fp.close()

        return profil_oce, file_list_pf_ocean


    def calc_bands(self, wl):
        '''
        calculate atot, btot at bands wl, and phase function at pfwav (or wl if
        pfwav is None)

        returns a list of (a, b, iphase) and a list of phase functions
        '''

        if (self.last is not None) and (self.last[0] == list(wl)):
            return (self.last[1], self.last[2])

        #
        # calculate the phase functions
        #
        if self.pfwav is None:
            pfwav = wl
        else:
            pfwav = self.pfwav

        phases_1 = []
        phases_2 = []
        phases_tot = []
        for w in pfwav:
            if self.verbose: print 'Calculate ocean phase function at', w

            _, [(b1, P1), (b2, P2)] = self.calc(w)

            phases_1.append(P1)
            phases_2.append(P2)
            ptot = (b1*P1 + b2*P2)/(b1+b2)
            phases_tot.append(ptot)

        #
        # calculate the IOPs
        #
        iops = []
        for w in wl:
            ipha = np.abs(w - np.array(pfwav)).argmin()

            a, [(b1, _), (b2, _)] = self.calc(w, skip_phase=True)

            # adjust the scattering coefficients according to the truncation coefficient
            b = b1*phases_1[ipha].coef_trunc + b2*phases_2[ipha].coef_trunc

            iops.append((a, b, ipha))

        self.last = (list(wl), iops, phases_tot)

        return iops, phases_tot





cdef class CLUT:

    # attributes
    cdef float[:] data
    cdef int ndim
    cdef float[:,:] axes   # (axis index, values)
    cdef long int [:,:] invax  # (axis index, indices)
    cdef int[:] shape

    cdef int[:] _index
    cdef int[:] _inf  # index of lower index (interpolation)
    cdef float[:] _x  # fraction for interpolation
    cdef int[:] _dim_interp # indices of the dimensions to interpolate
    cdef int[:] _interp      # whether dimension i is interpolated
    cdef int[:] dim_has_axis
    cdef float[:] scaling  # scaling factor for index lookup (N-1)/(V(N-1) - V(0))
    cdef int[:] clip   # per-axis behaviour for values lookup
    cdef float[:,:] bounds  # lower and upper bounds of each axis
    cdef int[:] reverse   # whether each axis is reversed
    cdef int debug

    # cdef methods
    cdef float get(self, int[:] x)
    cdef set(self, float value, int[:] x)
    cdef int index(self, int i, int j)
    cdef int indexf(self, int i, float x)
    cdef int lookup(self, int i, float v) except -999
    cdef float interp(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division

try:
    cfg = get_ipython()
    from ipywidgets import FloatProgress, Label, Box, Layout
    from IPython.display import display
    mode = 'notebook'
except NameError:
    try:
        from progressbar import ProgressBar, ETA, Percentage, Bar, FormatLabel
        mode = 'progressbar2'
    except ImportError:
        from progressbar import ProgressBar, Widget, ETA, Percentage, Bar
        mode = 'progressbar'


def Progress(vmax, activate=True):
    if not activate:
        return Progress_invisible()
    elif mode == 'notebook':
        return Progress_notebook(vmax)
    elif mode == 'progressbar2':
        return Progress_progressbar2(vmax)
    elif mode == 'progressbar':
        return Progress_progressbar(vmax)
    else:
        raise('Invalid mode '+mode)


class Progress_invisible(object):
    '''
    A progress bar that does nothing
    '''
    def update(self, value, message=''):
        pass
    def finish(self, message=''):
        pass


class Progress_notebook(object):

    def __init__(self, vmax):
        '''
        Initialize the progress bar object in the notebook

        vmax: maximum value of the progress bar
        '''
        self.vmax = vmax
        self.pbar = FloatProgress(min=0, max=vmax)
        self.label = Label()
        self.layout = Layout(display='flex', align_items='center')
        self.box = Box([self.pbar, self.label], layout=self.layout)
        display(self.box)

    def update(self, value, message=''):

        value = min(value, self.vmax)  # don't exceed max
        self.pbar.value = value
        self.label.value = message

    def finish(self, message=''):
        self.pbar.bar_style = 'success'
        self.pbar.value = self.vmax
        self.label.value = message


class Progress_progressbar2(object):

    def __init__(self, max):
        '''
        Initialize the progress bar objectusing library 'progressbar2'
        max: maximum value of the progress bar
        '''
        self.max = max
        self.label = FormatLabel('')
        self.pbar = ProgressBar(widgets=[self.label, ' ', Percentage(), Bar(), ETA()], max_value=max).start()

    def update(self, value, message=''):

        value = min(value, self.max)  # don't exceed max
        self.label.format = message
        self.pbar.update(value)

    def finish(self, message=''):
        self.pbar.finish()
        self.label.format = message

class Progress_progressbar(object):

    def __init__(self, max):
        '''
        Initialize the progress bar object using library 'progressbar'
        max: maximum value of the progress bar
        '''
        class Custom(Widget):
            def update(self, bar):
                try: return self.__text
                except: return ''
            def set(self, text):
                self.__text = text

        self.max = max
        self.custom = Custom()
        self.pbar = ProgressBar(widgets=[self.custom, ' ', Percentage(), Bar(), ETA()], maxval=max).start()

    def update(self, value, message=''):

        value = min(value, self.max)  # don't exceed max
        self.custom.set(message)
        self.pbar.update(value)

    def finish(self, message=''):
        self.pbar.finish()
        self.custom.set(message)



if __name__ == '__main__':
    from time import sleep
    p = Progress(100, activate=True)
    for i in range(100):
        sleep(0.01)
        p.update(i, '[{}] running'.format(i))
    p.finish('done')
